// BLE packet constraints
const ble_packet_size = 512			// Maximum size of a packet in bytes
const ble_packet_header_size = 6	// A packet starts with a header this size, in bytes
const ble_packet_data_size = ble_packet_size - ble_packet_header_size

// PPP
const ppp_seed_1 = '/* PPP SEED 1 */'
const ppp_seed_2 = '/* PPP SEED 2 */'

// The time we wait between bluetooth connection attempts
const ble_max_reconnect_interval = 60000

var devices = {}

class BLE extends EventEmitter {

	constructor () {
		super()

		this.foundDevices = []
		this.enabled = false
		this.scanning = false
		this._scanTimer = null
		this.init()
	}

	// Megpróbáljuk engedélyezni a bluetooth-t
	init () {
		try {
			bluetoothle.initialize((result) => {
				if (result.status === "enabled") {
					this.enabled = true

					// A bluetooth BEKAPCSOLT állapotba került
					this.emit('bluetoothOn')
				} else {
					this.enabled = false

					// A bluetooth KIKAPCSOLT állapotba került
					this.emit('bluetoothOff')
				}
			}, {
				request: true,
				statusReceiver: true
			})
		} catch (err) {
			console.log('Error occurred while trying to initialize bluetooth: ', err)
		}
	}

	// Státusz frissítése és lekérdezése
	isEnabled () {
		return new Promise ((resolve) => {
			bluetoothle.isEnabled((result) => {
				this.enabled = result.isEnabled
				resolve(result.isEnabled)
			})
		})
	}

	// Helyzetmeghatározás állapota
	isLocationEnabled () {
		return new Promise ((resolve) => {
			bluetoothle.isLocationEnabled((result) => {
				this.enabled = result.isLocationEnabled
				resolve(result.isLocationEnabled)
			})
		})
	}

	// Helyzetmeghatározás bekapcsolása
	requestLocation () {
		return new Promise ((resolve) => {
			bluetoothle.requestLocation((result) => {
				resolve(result.requestLocation)
			}, (error) => {
				console.log('Error occurred while trying to request location: ', err)
				resolve(false)
			})
		})
	}

	// ACCESS_COARSE_LOCATION engedély meglétének lekérdezése
	hasPermission () {
		return new Promise ((resolve) => {
			bluetoothle.hasPermission((result) => resolve(result.hasPermission))
		})
	}

	// ACCESS_COARSE_LOCATION engedély kérése
	requestPermission () {
		return new Promise ((resolve) => {
			bluetoothle.requestPermission((result) => resolve(result.requestPermission), resolve(false))
		})
	}

	// Keresés indítása
	async startScan (scan_time = null) {
		//console.log('startScan triggered')
		this.foundDevices = []

		// Android API 23-tól (6.0) további engedélyek szükségesek
		if (cordova.platformId == 'android' && parseFloat(device.version) >= 6) {

			// ACCESS_COARSE_LOCATION engedély kérése
			while (!await this.hasPermission()) {
				await this.requestPermission()
			}

			// Helyzetmeghatározás bekapcsolása
			while (!await this.isLocationEnabled()) {
				await this.requestLocation()
			}
		}

		// Keresés maximális hossza
		if (scan_time) {
			this._scanTimer = setTimeout(async () => {
				await this.stopScan()
			}, scan_time)
		}

		// A keresés indul
		bluetoothle.startScan(async (result) => {

			// Ha a keresés most indult el
		    if (result.status === "scanStarted") {
				this.emit('scanOn')
				return
			}

			// Ha már éppen találtunk egy eszközt
			if (result.status === "scanResult") {

		        if (!this.foundDevices.some((device) => { return device.address === result.address })) {
		            this.foundDevices.push(result)

					// Új okostáskát fedeztünk fel
					this.emit('deviceFound', result)
		        }
		    }

		}, (error) => {

			console.log(error)

		}, { services: [ 'AA004C1D-7565-49EA-82AD-8688245FBE81' ] })
	}

	// Keresés állapot frissítése és lekérdezése
	isScanning () {
		return new Promise ((resolve) => {
			bluetoothle.isScanning((result) => {
				this.scanning = result.isScanning
				resolve(result.isScanning)
			})
		})
	}

	// Keresés leállítása
	stopScan () {

		// Keresés időzítő leállítása, ha szükséges
		if (this._scanTimer !== null) {
			clearTimeout(this._scanTimer)
			this._scanTimer = null
		}

		// Keresés leállítása
		return new Promise((resolve, reject) => {
			bluetoothle.stopScan((result) => {
				if (result.status === 'scanStopped') {
					this.emit('scanOff')
					resolve()
				} else {
					reject()
				}
			}, (error) => {
				// Ha esetleg eleve nem futott a keresés, azt ne jelezzük hibának
				if (error.message == 'Not scanning') {
					resolve()
				} else {
					reject(error)
				}
			})
		})
	}

	// Pair the device (not supported by SMB software dependencies)
	/*
	bond (address) {
		return new Promise((resolve, reject) => {
			bluetoothle.bond((result) => {
				console.log(`${address} is ${result.status}`)

				if (result.status === 'bonded') resolve()
				if (result.status === 'unbonded') reject('bonding unsuccessful')

			}, (error) => {

				reject(error)

			}, { address })
		})
	}
	*/

	// Is the device paired (not supported by SMB software dependencies)
	/*
	isBonded (address) {
		return new Promise ((resolve, reject) => {
			bluetoothle.isBonded((result) => {
				resolve(result.isBonded)
			}, (error) => {
				reject(error)
			}, { address })
		})
	}
	*/

	// Is the device currently connected?
	isConnected (address) {
		return new Promise ((resolve, reject) => {
			bluetoothle.isConnected((result) => {
				resolve(result.isConnected)
			}, (error) => {
				reject(error)
			}, { address })
		})
	}

	// Was the device connected before? (If so, for a new connection, the old session has to be closed first!)
	wasConnected (address) {
		return new Promise ((resolve, reject) => {
			bluetoothle.wasConnected((result) => {
				resolve(result.wasConnected)
			}, (error) => {
				reject(error)
			}, { address })
		})
	}

	// Connect to a device (new session)
	connect (address, smb = null) {

		// This is the promise that will be handled if connection succeeds / fails (first time only)
		let pr_resolve = null
		let pr_reject = null
		let pr_done = false
		let promise = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Timeout
		let cancel = setTimeout(async () => {
			if (!pr_done) {
				pr_done = true
				try {
					await this.close(address)
					pr_reject('Connection timeout')
				} catch (error) {
					pr_reject('Connection timeout - and also failed to close: ' + error)
				}
			}
		}, 10000)

		// Start trying to connect
		bluetoothle.connect((result) => {
			// Handle
			if (result.status === 'connected') {
				console.log('Connected to ', result.address)
				if (!pr_done) {
					pr_done = true
					pr_resolve(result.status)
				}
			} else {
				console.log('Disconnected from ', result.address)
				if (!pr_done) {
					pr_done = true
					pr_reject(result.status)
				} else {
					// Notify SMB via callback (only after a successful initial connection)
					if (smb !== null) smb._ble_disconnected()
				}
			}
		}, (error) => {
			pr_done = true
			clearTimeout(cancel)

			pr_reject(error)
		}, {
			address: address,
			autoConnect: false
		})

		return promise
	}

	// Kapcsolat bontása
	close (address) {
		return new Promise((resolve, reject) => {
			bluetoothle.close((result) => {
				if (result.status === 'closed') {
					console.log('Closed: ', address)
					resolve(result)
				} else {
					reject(result)
				}
			}, reject, { address })
		})
	}

	// Szolgáltalások felfedezése (kapcsolat szükséges)
	discover (address) {
		return new Promise((resolve, reject) => {
			bluetoothle.discover((result) => {
				if (result.status === 'discovered') {
					resolve(result)
				} else {
					reject(result)
				}
			}, reject, {
				address,
	  			clearCache: true
			})
		})
	}

	/*
	encode (data) {
		return bluetoothle.bytesToEncodedString(bluetoothle.stringToBytes(JSON.stringify(data)))
	}

	decode (data) {
		return JSON.parse(bluetoothle.bytesToString(bluetoothle.encodedStringToBytes(data)))
	}
	*/

	write (address, data) {
		return new Promise ((resolve, reject) => {

			bluetoothle.write((result) => {
				if (result.status === 'written') {
					//console.log('Written: ', data)
					resolve(result)
				} else {
					console.log('FUNKY STUFF AT WRITING', result)
					reject()
				}
			}, (error) => {
				console.log('Error occurred while trying to write: ', error)
				reject()
			}, {
				address,
				service: "AA004C1D-7565-49EA-82AD-8688245FBE81",
				characteristic: "19F3B2BF-6A58-45EC-9B83-ED27F6AB11C7",
				value: data,
				type: null
			})
		})
	}

	read (address) {
		return new Promise ((resolve, reject) => {

			bluetoothle.read((result) => {
				if (result.status === 'read') {

					if (typeof(result.value) === 'string') {
						//console.log('chunk buffer received: ', bluetoothle.encodedStringToBytes(result.value))
						resolve(bluetoothle.encodedStringToBytes(result.value))
					} else {
						reject()
					}

				} else {
					reject()
				}
			}, (error) => {
				console.log('Error occured while trying to read: ', error)
				reject()
			}, {
				address,
				service: "AA004C1D-7565-49EA-82AD-8688245FBE81",
				characteristic: "19F3B2BF-6A58-45EC-9B83-ED27F6AB11C7",
			})

		})
	}

	subscribe (address, callback) {

		return new Promise((resolve, reject) => {
			bluetoothle.subscribe((result) => {

				// We have registered the subscription
				if (result.status === "subscribed") {
					console.log('subscribed')
					resolve()
					return
				}

				// We are receiving
				if (result.status === "subscribedResult") {
					//let msg = bluetoothle.bytesToString(bluetoothle.encodedStringToBytes(result.value))
					let msg = new TextDecoder().decode(bluetoothle.encodedStringToBytes(result.value))
					//console.log('Notification:', msg)
					callback(msg)
				}

			}, (error) => {

				console.log(address, error)
				reject(error)

			}, {
				address,
				service: "AA004C1D-7565-49EA-82AD-8688245FBE81",
				characteristic: "19F3B2BF-6A58-45EC-9B83-ED27F6AB11C7",
			})
		})
	}

	unsubscribe (address) {
		return new Promise((resolve, reject) => {
			bluetoothle.unsubscribe((result) => {
					if (result.status === 'unsubscribed') {
						resolve()
					} else {
						reject('Could not unsubscribe')
					}
				}, (error) => {
					reject(error)
				}, {
					address,
					service: "AA004C1D-7565-49EA-82AD-8688245FBE81",
					characteristic: "19F3B2BF-6A58-45EC-9B83-ED27F6AB11C7",
				}
			)
		})
	}
}

/* This is the list of SMBs that the phone is associated with. This list can be saved as and
 * restored from a json object
 */
class SMBList extends EventEmitter {

	constructor (ble) {
		super()

		// Reference to the BLE driver instance
		this._ble = ble
		// Object to hold all associated SMB instances
		this._index = {}

		// When the bluetooth is enabled, notify all SMBs so they can connect immediately
		this._ble.on('bluetoothOn', () => {
			this.forEach(smb => smb._bluetoothOn())
		})

		// When the bluetooth is disabled, notify all SMBs so they can sever their connections
		this._ble.on('bluetoothOff', () => {
			this.forEach(smb => smb._bluetoothOff())
		})
	}

	// Deserialize from a json and restore previous settings
	restore (ble, save) {
		save.forEach((smb) => {
			this.add('SMB-' + smb.sn, smb.addr, smb.pin, smb.state )
		})
	}

	// Try to add a new SMB to the list
	add (name, addr, pin, state) {
		console.log('Adding bag:', name, addr, pin, state)

		// Check whether or not the SMB is already on the list..
		let sn = name.substring(4, 12)
		if (this._index.hasOwnProperty(sn + addr)) {
			console.log('That bag is already on the list, likely in disconnected state')
			return
		}

		// ..it's not, so create it
		let smb = new SMB(sn, addr, pin, this, state)


		// Persistent data
		/*
		if (beacons !== null) {
			smb._state.b = beacons
		}
		*/

		// All done, we can add it to the list and notify now
		this._index[sn + addr] = smb
		this.emit('smb-added', smb)
	}

	// Remove an SMB instance and notify the UI
	async remove (sn, addr) {

		console.log('remove started: ' + sn + addr)

		if (!this._index.hasOwnProperty(sn + addr)) return

		// A self destruct call for the SMB class, we cannot rely on the gc.
		await this._index[sn + addr].destructor()

		// Remove from the list
		delete this._index[sn + addr]

		console.log('smb removed, event is about to trigger')

		this.emit('smb-removed', {sn, addr})
	}

	// Get the SMB with the given serial number and mac address
	get (sn, addr) {
		return this._index[sn + addr]
	}

	// Get the SMB occupying the given offset in the list
	getAt (offset) {
		let keys = Object.keys(this._index)
		if (offset >= keys.length) return null
		return this._index[keys[offset]]
	}

	// Get the length of the SMB List
	length () {
		return Object.keys(this._index).length
	}

	// Iterate over all the SMBs
	forEach (cb) {
		Object.values(this._index).forEach(cb)
	}

	// Serialize into a json for saving
	serialize () {

		let save = []
		this.forEach((smb) => {

			// Save all the beacons that are linked
			let linked_beacons = { }
			Object.keys(smb._state.b).forEach((addr) => {
				if (smb._state.b[addr].p == 1) {
					linked_beacons[addr] = smb._state.b[addr]
				}
			})

			save.push({
				sn: smb.serial_nr,
				addr: smb.ble_address,
				pin: smb.pin,
				state: smb._state
			})
		})
		return save

	}
}

/* This class represents a Smart Modding Bag (SMB). All interaction with the SMB is handled through
 * the methods and events of this object.
 */
class SMB extends EventEmitter {

	constructor (serial_nr, ble_address, pin, list, state) {
		super()

		console.log('SMB being constructed', serial_nr, ble_address, state)

		this._serial_nr = serial_nr
		this._pin = pin
		this._state = {
			w: {
				k: ''
			},
			h: {
				e: 0,			// hostpot enabled
				s: '',			// currently used ssid
				p: ''			// currently used password
			},
			l: {
				a: 0,			// latitude
				o: 0,			// longitude
				t: 0,			// timestamp (in bag time)
				s: 0			// location source: 0 - gps, 1 - gsm
			},
			v: {				// battery
				p: null			// percentage
			},
			b: { }				// beacons
		}

		if (typeof state != 'undefined') this._state = state

		this._destroyed = false

		// Reference to the BLE driver
		this._ble = list._ble

		// API related
		this._requests = {}
		this._request_count = 0
		this._smb_connected = false
		this._smb_connection_change_stamp = null
		this._time_offset = 0

		// Bluetooth connection
		this._ble_address = ble_address
		this._ble_status = 'disconnected'	// disconnected | connecting | handshake | authorized
		this._ble_device_info = null
		this._ble_reconnect_timer = null
		this._ble_reconnect_interval = 0

		this._ble_ps_count = 0	// Number of payloads sent
		this._ble_pr_count = 0	// Number of payloads received
		this._ble_queue = []
		this._ble_processing = false

		// Internet connection
		//this._ppp_channel_id = (this._serial_nr + this._ble_address.split(':').join('') + md5(ppp_seed_2 + md5(ppp_seed_1 + this._pin)).substring(0, 16)).toUpperCase()
		this._ppp_channel_id = (this._serial_nr + md5(ppp_seed_2 + md5(ppp_seed_1 + this._pin)).substring(0, 28)).toUpperCase()
		this._ppp_smb_id = this._ppp_channel_id + '-1'
		this._ppp_app_id = this._ppp_channel_id + '-2'
		this._ppp_status = 'disconnected'	// disconnected | connecting | handshake | authorized

		this._ppp_connected = false
		this._ppp_client = new PPP_Client(this._ppp_app_id, this._ppp_smb_id)

		this._ppp_client.on('state-change', (pppstate) => {
			this._ppp_connected = pppstate
			console.log(this._ppp_connected ? 'CONNECTED TO REMOTE DEVICE' : 'DISCONNECTED FROM REMOTE DEVICE')

			this._ppp_state = this._ppp_connected ? 'authorized' : 'disconnected'

			if (this._ppp_connected) this.sync()
		})

		this._ppp_client.on('receive', (payload) => {
			// Sort out what to do with the message
			this._handle(payload) // <- This is async, but we are not waiting for it
		})

		console.log('SMB:' + this._serial_nr + ' channel id:' + this._ppp_app_id)

		// Reference to the SMB list
		this._list = list

		// Let's try to connect
		this._ble_connect()
		/* This is async, but we are not waiting for it. This way, the constructor finishes, and
		 * allows the SMB onto the list before it connects.
		 */
	}

	async destructor () {

		console.log('Destructing SMB: ', this._serial_nr)

		this._destroyed = true

		// Remove event handlers
		this.events = []

		// Close connection
		try {
			await this._ble.close(this._ble_address)
		} catch (err) { }

		// Stop all timers
		clearTimeout(this._ble_reconnect_timer)
		delete this._ble_reconnect_timer

		// Remove everything
		for (let key in this) {
			if (this.hasOwnProperty(key)) delete this[key]
		}
	}

	// API

	get ble_address () { return this._ble_address }
	get serial_nr () { return this._serial_nr }
	get pin () { return this._pin }
	get ble_state () { return this._ble_status }
	get ppp_state () { return this._ppp_status }
	get online_state () {
		return this._smb_connected
	}
	get last_seen () {
		if (this.online_state == true) {
			return Date.now()
		}
		return this._smb_connection_change_stamp
	}
	get time_offset () { return this._time_offset }
	get battery_charge () { return this._state.v.p }
	get location () {
		return { lat: this._state.l.a, lon: this._state.l.o, time: this._state.l.t, source: ['gps', 'gsm'][this._state.l.s] }
	}
	get hotspot () {
		return { enabled: (this._state.h.e == 1), ssid: this._state.h.s, password: this._state.h.p }
	}
	get client_key () { return this._state.w.k }

	get beacons () { return this._state.b }

	// Ping the bag
	ping (cb) {

		// Get a request ID
		let request_id = ++this._request_count

		// Mark the time
		let start = Date.now()

		// Make the request object
		let request = {
			i: request_id,
			r: 0
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error callback
			cb('Timeout')
		}, 10000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			// Call cb with elapsed time
			cb(null, Date.now() - start)
		}

		// Send it
		this._send(request)
	}

	// Request a full status object
	sync () {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 2
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('sync failed', this._ble_address)
			pr_reject('sync failed ' + this._ble_address)
		}, 10000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			console.log('sync reply arrived', reply.s)
			this._state = reply.s
			pr_resolve()
			this.emit('smb-mutated')
		}

		// Send it
		this._send(request)

		return pr
	}

	// Link a beacon
	linkBeacon (address, name = '', type = 0, essential = false) {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 5,
			a: address,
			n: name,
			t: type,
			e: essential
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('beacon link failed', this._ble_address)
			pr_reject('link failed ' + this._ble_address)
		}, 10000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			if (reply.s == 1) {
				if (this._state.b.hasOwnProperty(address)) {
					this._state.b[address].p = 1
					this._state.b[address].n = name
					this._state.b[address].t = type
					this._state.b[address].e = essential
				}

				pr_resolve()
				this.emit('smb-mutated')
			} else {
				pr_reject('link failed ' + this._ble_address)
			}
		}

		// Send it
		this._send(request)

		return pr
	}

	// Unlink a beacon
	unlinkBeacon (address) {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 6,
			a: address
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('beacon unlink failed', this._ble_address)
			pr_reject('unlink failed ' + this._ble_address)
		}, 10000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			if (reply.s == 1) {
				if (this._state.b.hasOwnProperty(address)) {
					this._state.b[address].p = 0
				}
				pr_resolve()
				this.emit('smb-mutated')
			} else {
				pr_reject('unlink failed ' + this._ble_address)
			}
		}

		// Send it
		this._send(request)

		return pr
	}

	// Alias for update beacon
	setBeacon (...params) {
		this.updateBeacon(...params)
	}

	// Update a beacon's settings
	updateBeacon (address, name = '', type = 0, essential = false) {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 7,
			a: address,
			n: name,
			t: type,
			e: essential
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('beacon update failed', this._ble_address)
			pr_reject('beacon update failed ' + this._ble_address)
		}, 10000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			if (reply.s == 1) {
				if (this._state.b.hasOwnProperty(address)) {
					this._state.b[address].p = 1
					this._state.b[address].n = name
					this._state.b[address].t = type,
					this._state.b[address].e = essential
				}
				pr_resolve()
				this.emit('smb-mutated')
			} else {
				pr_reject('beacon update failed ' + this._ble_address)
			}
		}

		// Send it
		this._send(request)

		return pr
	}

	// Enable / disable hotspot
	setHotspot (enable = true, ssid = null, pass = null) {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 14,
			e: (enable ? 1 : 0),
			s: ssid,
			p: pass
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('setting hotspot failed', this._ble_address)
			pr_reject('setting hotspot failed ' + this._ble_address)
		}, 13000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			if (reply.s == 1) {
				pr_resolve()
				this.emit('hotspot-set')
			} else {
				pr_reject('setting hotspot failed ' + this._ble_address)
			}
		}

		// Send it
		this._send(request)

		return pr
	}

	resetClientKey () {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 15
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('resetting client key failed', this._ble_address)
			pr_reject('resetting client key failed ' + this._ble_address)
		}, 30000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			if (reply.s == 1) {
				this._state.w.k = reply.k
				pr_resolve(reply.k)
				this.emit('client-key', reply.k)
			} else {
				pr_reject('resetting client key failed ' + this._ble_address)
			}
		}

		// Send it
		this._send(request)

		return pr
	}

	// Sort out what to do with incoming messages
	async _handle (request) {

		// Ignore anything that is not an object
		if (typeof request !== 'object') return

		// Ignore any requests without a type identifier
		if (!request.hasOwnProperty('r')) return

		// Let's see what type of request we are dealing with, and handle accordingly
		switch (request.r) {

			/* Request ids:
			 * 0 -> auth
			 * 1 <- ping
			 */

			case 1: // ping
			case 2: // sync
			case 5: // beacon link
			case 6: // beacon unlink
			case 7: // update beacon
			case 10: // auth challenge and timesync
			case 11: // auth response
			case 14: // hotspot set
			case 15: // reset client key

				// This is a reply, ignore if no corresponding request is pending anymore
				if (!request.hasOwnProperty('i') || !this._requests.hasOwnProperty(request.i)) break
				// Let the request handle it's reply
				this._requests[request.i](request)
				break

			case 3: // beacon update
				this._state.b[request.k] = request.v
				this.emit('beacon-update', request.k)
				this.emit('smb-mutated')
				break

			case 4: // beacon remove
				delete this._state.b[request.k]
				this.emit('beacon-remove', request.k)
				this.emit('smb-mutated')
				break

			case 12: // battery voltage update
				this._state.v.p = request.p
				this.emit('battery-charge', this.battery_charge)
				break

			case 13: // location update
				this._state.l.a = request.a
				this._state.l.o = request.o
				this._state.l.t = request.t
				this._state.l.s = request.s
				this.emit('location', this.location)
				break

			case 16: // hotspot state
				this._state.h.e = request.e
				this._state.h.s = request.s
				this._state.h.p = request.p
				this.emit('hotspot-state', this.hotspot)
				break

			default: // any unimplemented requests..
				console.log('Received unimplemented request', request)	// Just print it
				break
		}

	}

	// Internals -----------------------------------------------------------------------------------

	// === General networking

	// Put a message object on the send queue
	_send (data, force_ble = false) {
		if (this._ppp_status !== 'disconnected' && !force_ble) {
			this._ppp_client.send_message(data)
		} else if (this._ble_status !== 'disconnected') {
			this._ble_queue.push(BlePayload.Outgoing(data))
			this._ble_process()
		} else {
			console.log('Unable to send (not connected): ', data)
		}
	}

	// === BLE specific networking

	async _ble_connect () {

		// Mark the timer completed
		if (this._ble_reconnect_timer !== null) {
			clearTimeout(this._ble_reconnect_timer)
			this._ble_reconnect_timer = null
		}

		if (this._ble_status != 'disconnected') {
			console.log('There was a _ble_connect call, but the SMB is NOT DISCONNECTED:', this._ble_address)
			// Do not attempt to connect if the SMB is not in a disconnected state
			return
		}

		if (this._ble.enabled == false) {
			console.log('There was a _ble_connect call, but the BLE DEVICE is NOT READY yet:', this._ble_address)
			// Do not attempt to connect if the device is not ready yet
			this._ble_state = 'disconnected'
			return
		}

		try {
			// Attempt to connect ------------------------------------------------------------------
			console.log(this._ble_address + ' connecting')
			this._ble_state = 'connecting'

			if (await this._ble.wasConnected(this._ble_address)) {
				if (await this._ble.isConnected(this._ble_address)) {
					console.log(1)
					// Already connected
				} else {
					console.log(2)
					// Disconnected, so wrap up and start over as suggested in the bluetoothle docs
					await this._ble.close(this._ble_address)
					await this._ble.connect(this._ble_address, this)
				}
			} else {
				console.log(3)
				// Not connected yet, so do so now
				await this._ble.connect(this._ble_address, this)
			}

			// Handshake ---------------------------------------------------------------------------
			console.log(this._ble_address + ' handshake')
			this._ble_state = 'handshake'

			// Run discovery to gain access to services
			this._ble_device_info = await this._ble.discover(this._ble_address)
			console.log(this._ble_address + ' discovered')

			// Subscribe to notifications so we can "receive" packets
			await this._ble.subscribe(this._ble_address, (msg) => { this._ble_notify(msg) })
			console.log(this._ble_address + ' subscribed')

			// Auth
			let challenge = await this._ble_auth_challenge()
			console.log('challenge received')
			let response = md5(challenge + this._pin)
			let result = await this._ble_auth_response(response)
			console.log('response sent')
			if (result) {
				// Auth success
			} else {
				// Auth fail
				this.emit('auth-fail')
				await this._list.remove(this._serial_nr, this._ble_address)
				return
			}

			// Stage complete
			this.emit('auth-success')

			// Sync
			await this.sync()

			// Authorized --------------------------------------------------------------------------
			console.log(this._ble_address + ' authorized')
			this._ble_state = 'authorized'

			// Debug - send all of this at once
			/*
			this._send({test:'Hello world!'})
			this._send('ping')
			this._send({test:'654631686516846516846516546544539864654546648566354153648576346684463244365'})
			this._send({test:'6546316865168465168465165465445398654546648566354153648576346684463244365'})
			this._send({test:'65463168651684651684651654654453954546648566354153648576346684463244365'})
			this._send({
				test: 'sdéfohsrt lshgy lduigh yldgh ylrejghy ldrighyerg yuhdrl gjydhgl iyuhg,ykjgh ylbjgykidughykjgbyrgjybrgkjyhrgyjhglyiuerhlydjhgylieurthykj bs5eit8zhltsuihe89se5t hseuith sp958tzsh.etuog shp958tzalsz le5t8sh5l s8z e5ltzisuh5 zly8ieg she.jy éerg8 z5lt usyhrt gp98e5zuseél8tz leshe85tz ',
				test2: 'sdéfohsrt lshgy lduigh yldgh ylrejghy ldrighyerg yuhdrl gjydhgl iyuhg,ykjgh ylbjgykidughykjgbyrgjybrgkjyhrgyjhglyiuerhlydjhgylieurthykj bs5eit8zhltsuihe89se5t hseuith sp958tzsh.etuog shp958tzalsz le5t8sh5l s8z e5ltzisuh5 zly8ieg she.jy éerg8 z5lt usyhrt gp98e5zuseél8tz leshe85tz ',
				test3: 'sdéfohsrt lshgy lduigh yldgh ylrejghy ldrighyerg yuhdrl gjydhgl iyuhg,ykjgh ylbjgykidughykjgbyrgjybrgkjyhrgyjhglyiuerhlydjhgylieurthykj bs5eit8zhltsuihe89se5t hseuith sp958tzsh.etuog shp958tzalsz le5t8sh5l s8z e5ltzisuh5 zly8ieg she.jy éerg8 z5lt usyhrt gp98e5zuseél8tz leshe85tz ',
				test4: true,
				test5: 3453645,
				test6: null
			})
			*/

		} catch (error) {

			// Close connection
			try {
				await this._ble.close(this._ble_address)
			} catch (err) {
				console.log('Could not close connection after a connection attempt failure', err)
			}

			// At any point during the connection, something went wrong
			console.log(this._ble_address + ' disconnected')
			this._ble_state = 'disconnected'

			console.log('Connection failed: ', error)
		}
	}

	// If we are not connected and the BLE device just came on, we attempt to connect
	async _bluetoothOn () {
		if (this._ble.enabled && this._ble_status == 'disconnected') {
			// This will reset the reconnection timer automatically
			this._ble_connect()
		}
	}

	async _bluetoothOff () {

		if (this.ble_state !== 'disconnected') {

			// Close connection
			/*
			try {
				await this._ble.close(this._ble_address)
			} catch (err) {
				console.log('Could not close connection', err)
			}
			*/

			// Mark closed
			this._ble_state = 'disconnected'
		}
	}

	// Private state property, emits an event every time the status changes
	set _ble_state (value) {

		if (this._destroyed) return

		let old_status = this._ble_status
		this._ble_status = value
		this.emit('ble-state-change', this._ble_status)

		// Global connectivity state
		let connected = ( this._ble_status == 'authorized' || this._ppp_status == 'authorized' )
		if (this._smb_connected !== connected) {
			this._smb_connected = connected
			this._smb_connection_change_stamp = Date.now()
			this.emit('online-state-change', this._smb_connected)
		}

		/* If for *any* reason the connection is closed, we schedule a reconnection attempt after a
		 * period of ble_max_reconnect_interval milliseconds. Each connection attempt will reset the
		 * connection status, in effect restarting the timer when needed.
		 */
		if (value == 'disconnected') {
			// We are disconnected, so we should have a timer..
			if (this._ble_reconnect_timer == null) {

				/* Determine the timeout to wait
				 * If we have just disconnected, start with a 10 second delay. Then, gradually
				 * increase the time after each try, until we reach ble_max_reconnect_interval.
				 */
				if (old_status == 'authorized') {
					// This means that we have just disconnected, so let's make the delay short
					this._ble_reconnect_interval = 0
				}
				if (this._ble_reconnect_interval < ble_max_reconnect_interval) {
					this._ble_reconnect_interval += 10000
				}

				this._ble_reconnect_timer = setTimeout(this._ble_connect.bind(this), this._ble_reconnect_interval)
				console.log('reconnect timer started for ' + this._ble_address + ' ' + this._ble_reconnect_interval + 'ms')
			}
		} else {
			// We are now in a different state, so make sure we have the timer cleared
			if (this._ble_reconnect_timer !== null) {
				clearTimeout(this._ble_reconnect_timer)
				this._ble_reconnect_timer = null
				console.log('reconnect timer cleared for ' + this._ble_address)
			}
		}
	}

	set _ppp_state (value) {

		if (this._destroyed) return

		this._ppp_status = value
		this.emit('ppp-state-change', this._ppp_status)

		// Global connectivity state
		let connected = ( this._ble_status == 'authorized' || this._ppp_status == 'authorized' )
		if (this._smb_connected !== connected) {
			this._smb_connected = connected
			this._smb_connection_change_stamp = Date.now()
			this.emit('online-state-change', this._smb_connected)
		}
	}

	// Start auth sequence
	_ble_auth_challenge () {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 10
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('auth failed', this._ble_address)
			pr_reject('auth failed ' + this._ble_address)
		}, 10000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			// Time offset between local and SMB time
			this._time_offset = reply.t - Date.now()

			// Return with the challenge string
			pr_resolve(reply.c)
		}

		// Send it
		this._send(request, true)

		return pr
	}

	// Finish auth sequence
	_ble_auth_response (response) {

		// Get a request ID
		let request_id = ++this._request_count

		let pr_resolve = null
		let pr_reject = null
		let pr = new Promise((resolve, reject) => {
			pr_resolve = resolve
			pr_reject = reject
		})

		// Make the request object
		let request = {
			i: request_id,
			r: 11,
			a: response
		}

		// Define a timeout for the reply
		let timeout = setTimeout(() => {
			delete this._requests[request_id]

			// Error
			console.log('auth failed', this._ble_address)
			pr_reject('auth failed ' + this._ble_address)
		}, 10000)

		// Define how to handle the reply
		this._requests[request_id] = (reply) => {
			clearTimeout(timeout)

			if (reply.s == 1) {
				// Auth success
				pr_resolve(true)
			} else {
				// Auth fail
				pr_resolve(false)
			}
		}

		// Send it
		this._send(request, true)

		return pr
	}

	// Callback for the BLE class to call when the connection status changes
	_ble_disconnected () {
		this._ble_state = 'disconnected'

		// Handle any tasks after disconnecting?
		// ...
	}

	// Process the queue: send or receive payloads
	async _ble_process () {

		// Make sure this only happens in this thread
		if (this._ble_processing) return
		this._ble_processing = true

		// Handle all pending transmissions
		while (this._ble_queue.length > 0) {

			let payload = this._ble_queue.shift()
			if (payload.mode == 'out') {
				// Send this payload
				//console.log('sending payload')
				await this._ble_send(payload)
			} else {
				// Read this payload
				//console.log('receiving payload')
				await this._ble_receive(payload)
			}
		}

		// Release
		this._ble_processing = false
	}

	// Receive a notification from the SMB
	_ble_notify (msg) {
		let params = msg.split(',')
		if (params.length > 0) {
			switch (params[0]) {
				case 'r': // SMB tells us that we should read a payload from it
					if (params.length < 2) {
						// missing parameter - ignore (TODO: handle error on SMB with a timeout)
						break
					}
					this._ble_queue.push(BlePayload.Incoming(parseInt(params[1])))
					this._ble_process()
				break
			}
		}
	}

	// Send a message to the SMB
	async _ble_send (payload) {

		const series_nr = ++this._ble_ps_count // Payload series number

		if (payload.length < 1) throw 'Data is empty.'
		if (payload.length > 255) throw 'Data is too long.'

		// Slice the buffer to packets no longer than ble_packet_size bytes
		for (let i = 0; i < payload.length; i++) {

			let chunk = payload.chunk(i)

			// Create the packet
			let packet = new ArrayBuffer(ble_packet_header_size + chunk.length),
				packet_dv = new DataView(packet)

			// Write header ----------------------
			// Packet series number (32 bit big endian)
			packet_dv.setUint32(0, series_nr, false)
			// Packet number (8 bit)
			packet_dv.setUint8(4, i + 1)
			// Packet count (8 bit)
			packet_dv.setUint8(5, payload.length)

			// Copy chunk data -------------------
			for (let p = 0; p < chunk.length; p++) {
				packet_dv.setUint8(ble_packet_header_size + p, chunk[p])
			}

			let payload_bytes = new Uint8Array(packet)
			//console.log('Compiled chunk: ', series_nr, i)

			// Write encoded bytes
			await this._ble.write(this._ble_address, bluetoothle.bytesToEncodedString(payload_bytes))
		}
	}

	// Read a chunk of data from the SMB
	async _ble_receive (payload) {

		//console.log('receiving........')
		const series_nr = ++this._ble_pr_count // Payload series number

		// Read data -------------------
		for (let i = 0; i < payload.length; i++) {

			let packet = await this._ble.read(this._ble_address)
			//let chunk = packet.subarray(ble_packet_header_size)

			// Validate the packet header
			// TODO

			// Add the chunk data to the payload
			payload.store(i, packet)
		}

		// Make sure that the payload has been completed
		if (!payload.is_complete) {
			throw 'read error: incomplete payload'
		}

		//console.log('Received: ', payload.value)

		// Sort out what to do with the message
		this._handle(payload.value) // <- This is async, but we are not waiting for it
		return
	}
}

/* A payload represents a single message sent one way over bluetooth. It will either be an incoming
 * or outgoing message. The payload may be split into packets of a specific max size, so it can be
 * transferred over BLE.
 */
class BlePayload {

	static Incoming (length) {
		let payload = new BlePayload('in')
		payload._length = length
		payload._series = new Array(payload._length)
		return payload
	}

	static Outgoing (value) {
		let payload = new BlePayload('out')
		payload.value = value
		return payload
	}

	constructor (mode) {
		this._mode = mode
		this._time = Date.now()
		this._length = null
		this._series = null			// Array of buffers that contain chunks of the payload
	}

	// Is this an incoming or an outgpoing payload?
	get mode () { return this._mode }

	// How many packets are expected in this series?
	get length () { return this._length }

	// Get a chunk of binary data
	chunk (i) { return this._series[i] }

	// Add (or replace) chunk data (incoming payloads)
	store (offset, buffer) {
		this._series[offset] = buffer
	}

	// Check whether or not all the chunks are present (incoming payloads)
	is_complete () {
		for (let i = 0; i < this._length; i++) {
			if (typeof this._series[i] === 'undefined')
				return false
		}
		return true
	}

	// Return the decoded payload object
	get value () {
		// Convert the series into the message object
		try {
			if (!this.is_complete())
				throw 'data incomplete'

			// Combine the strings encoded in the chunks
			let str = ''
			for (let i = 0; i < this._length; i++) {
				str += new TextDecoder().decode(this._series[i])
			}

			// Return the original json
			return JSON.parse(str)

		} catch (error) {
			throw 'invalid payload - ' + error
		}
	}

	// Set the payload object
	set value (value) {
		// Convert the message object into a series, and write it into a memory buffer
		//let buffer = bluetoothle.stringToBytes(JSON.stringify(value))
		let buffer = new TextEncoder().encode(JSON.stringify(value))

		// Slice the buffer to chunks no longer than ble_packet_data_size bytes
		this._length = Math.ceil(buffer.length / ble_packet_data_size)
		this._series = new Array(this._length)
		for (let i = 0; i < this._length; i++) {
			this._series[i] = buffer.subarray(i * ble_packet_data_size, (i + 1) * ble_packet_data_size)
		}
	}
}

class PPP_Client extends EventEmitter {

	constructor (me, partner) {
		super()

		console.log('Initializing socket connection')

		this._me = me
		this._partner = partner

		this._active = false
		this._connected = false
		this._socket = null
		this._seq = 0
		this._heartbeat_interval = null
		this._last_incoming = Date.now()

		this.connect()
	}

	connect () {

		console.log(this._me, 'TUNNEL - Attempting to open socket')

		if (this._active) {
			console.log('Cannot open socket, because it is already active')
			return
		}

		// Recreate socket
		this._active = true
		this._socket = new WebSocket("/* YOUR SMB WS URL */")

		// Socket opened
		this._socket.onopen = e => {
			console.log(this._me, 'TUNNEL Socket Open')

			// Start heartbeats
			this._heartbeat_interval = setInterval(this.heartbeat.bind(this), 15000)

			this.send({
				type: "open",
				me: this._me,
				rec: this._partner
			})
		}

		this._socket.onerror = (e) => {
			console.log(this._me, "TUNNEL Socket Error", e)
		}

		this._socket.onclose = (e) => {
			console.log(this._me, "TUNNEL Socket Closed", e)

			// Stop heartbeats
			clearInterval(this._heartbeat_interval)
			this._heartbeat_interval = null

			this.onclose()
		}

		this._socket.onmessage = (e) => {
			console.log(this._me, "Message", e.data)

			let payload = JSON.parse(e.data)

			this._last_incoming = Date.now()

			if (!payload.hasOwnProperty('type')) return
			switch (payload.type) {
				case 'ack':
				case 'control':
					// We received a control or ack message, update the connection state
					let old_connection_state = this._connected
					this._connected = (payload.m.cli == 2)

					// If the state changed, let's notify
					if (this._connected !== old_connection_state) {
						this.emit('state-change', this._connected)
					}

					break
				case 'data':
					// We received a data message
					console.log('TUNNEL RECEIVE', payload.d)
					this.emit('receive', payload.d)
					break
			}
		}
	}

	heartbeat () {

		// Connection timeout
		if (this._last_incoming < Date.now() - 30000) {
			this._socket.close()
			return
		}

		// Keepalive
		if (this._socket.readyState == 1) {
			console.log(this._me, 'TUNNEL Heartbeat')
			this.send({ type: "ping" })
		}
	}

	send_message (data) {
		return this.send({
			type: "send",
			d: data,
		})
	}

	send (payload) {
		payload.seq = ++this._seq
		this._socket.send(JSON.stringify(payload))
		return payload.seq
	}

	close () {
		if (this._socket !== null) {
			// Close the socket
			this._socket.close()
		} else {
			// The socket was never open to begin with
			this.onclose()
		}
	}

	// This function will always be called when the connection is lost / closed
	async onclose () {
		this._active = false

		let old_connection_state = this._connected
		this._connected = false

		// If the state changed, let's notify
		if (this._connected !== old_connection_state) {
			this.emit('state-change', this._connected)
		}

		// Wait a few seconds
		await new Promise((resolve) => { setTimeout(resolve, 5000) })

		// Reconnect
		this.connect()
	}
}
