'use strict';
const CONST = require("../constants.js");
const LOG   = require("../core/Logger.js");

const ServiceBase    = require("../core/ServiceBase.js");

const MongoConnector         = require("./dal/MongoConnector.js");
const DataAccessLayer = require("./dal/DataAccessLayer.js");

const FileStorageConnector = require("./storage/FileStorageConnector.js");
const BlobStorage          = require("./storage/BlobStorage.js");

const DBSCAN = require("./DBSCAN.js");

class ProcessorService extends ServiceBase {
	_initialize(done) {
		this._setupConnector();
		this._setupDataAccessLayer();

		this._setupBlobStorageConnector();
		this._setupBlobStorage();

		this._requestCnt = 0;
		done();
	}

	_setupBlobStorageConnector() {
		let permission = parseInt(this._config.get("blob.permission", "660"), 8);
		let encoding   = this._config.get("blob.encoding", "utf8");

		this._blobConnector = new FileStorageConnector(permission, encoding);
	}

	_setupBlobStorage() {
		let rootPath  = this._config.get("blob.root");
		let hashDepth = this._config.get("blob.hash_depth", 3);
		this._blobStorage = new BlobStorage(this._blobConnector, rootPath, hashDepth);
	}

	_setupConnector() {
		this._connector = new MongoConnector(
			this._config.get("dal.url", "mongodb://localhost"),
			this._config.get("dal.db", "smb")
		);
	}

	_setupDataAccessLayer() {
		this._dal = new DataAccessLayer(this._connector);
	}

	_buildDataVector(data) {
		let ret = [];
		for (let i in data) {
			let record = data[i];
			let lines  = record.split("\n");
			for (let i in lines) {
				try {
					let line = JSON.parse(lines[i]);
					if (line.lon && line.lat) {
						ret.push([line.lon, line.lat])
					}
				} catch (e) {}
			}
		}

		return ret;
	}

	_buildClusters(dataVector) {
		let clust = new DBSCAN(dataVector);
		let clusterMap = clust.buildClusters(100, 0.0001);

		let clusters = {};
		for (let i in dataVector) {
			let vector = dataVector[i];
			if (!clusters[clusterMap[i]]) {
				clusters[clusterMap[i]] = {
					sumVector : [],
					cnt       : 0,
					rr        : -1
				}
			}
			let cluster = clusters[clusterMap[i]];
			for (let j = 0; j < vector.length; j++) {
				cluster.sumVector[j] = ((cluster.sumVector[j] || 0) * cluster.cnt + vector[j]) / (cluster.cnt + 1);
			}
			cluster.cnt++;
		}

		for (let i in dataVector) {
			let vector = dataVector[i];
			let cluster = clusters[clusterMap[i]];
			let d = 0;
			for (let j = 0; j < vector.length; j++) {
				d += (cluster.sumVector[j] - vector[j]) * (cluster.sumVector[j] - vector[j]);
			}
			if (d > cluster.rr) {
				cluster.rr = d;
				cluster.r  = Math.sqrt(d);
			}
		}

		//removing dummy clusters
		delete clusters[-1];
		delete clusters[0];

		let clusterArr = [];
		for (let i in clusters) {
			clusterArr.push({
				lat : clusters[i].sumVector[1],
				lon : clusters[i].sumVector[0],
				rr  : clusters[i].rr,
				r   : clusters[i].r,
				cnt : clusters[i].cnt
			});
		}
		clusterArr.sort((a,b) => b.cnt - a.cnt).splice(10);

		return clusterArr;
	}

	_updateBagData(bagId, safePlaces) {
		return this._dal.getBagData(bagId)
			.then(bag => {
				bag.setSafePlaces(safePlaces);
				return this._dal.updateBagData(bagId, bag);
			})
	}

	async _processBag(id) {
		LOG.info("Processing", id);
		return this._blobStorage.getBagLogs(id)
			.then(r => this._buildDataVector(r))
			.then(r => this._buildClusters(r))
			.then(r => this._updateBagData(id, r));
	}

	async _processList(list) {
		LOG.info("Found ", list.length," bags, starting to process");
		let tasks = [];
		for (let i in list) {
			let item = list[i]._id;
			await this._processBag(item).catch(e => LOG.error("Failed to process", item));
		}
	}

	_procession() {
		return this._connector.listRecords("bagData")
			.then(r => this._processList(r));
	}

	_run() {
		LOG.info("Connecting to database");
		this._connector
			.connect()
			.then(r => this._procession())
			.catch(e => {
				LOG.error("Error during startup", e);
			})
			.then(e => this._terminate());
	}

	_terminate(done) {
		this._connector.close();
	}
}

module.exports = ProcessorService;
