'use strict';
const StorageConnectorBase = require("./StorageConnectorBase.js");
const StorageException     = require("./StorageException.js");

const fs  = require('fs');
const fsp = fs.promises;

class FileStorageConnector extends StorageConnectorBase {
	constructor(permission, encoding) {
		super();
		this._permission = permission || 0o640;
		this._encoding   = "utf8";
	}

	_errorHandler(e) {
		let code = 0;

		switch (e.code) {
			case "ENOENT":
				code = StorageException.ERROR_NOT_FOUND;
			break;
			case "EPERM":
			case "EACCES":
				code = StorageException.ERROR_PERMISSION_DENIED;
			break;
			case "EEXIST":
				code = StorageException.ERROR_ALREADY_EXIST;
			break;
			case "ENOTEMPTY":
				code = StorageException.ERROR_NOT_EMPTY;
			break;
			case "EISDIR":
				code = StorageException.ERROR_NOT_A_FILE;
			break;
			case "ENOTDIR":
				code = StorageException.ERROR_NOT_A_DIRECTORY;
			break;
			case "EPIPE":
				code = StorageException.ERROR_BROKEN_PIPE;
			break;
			case "EMFILE":
				code = StorageException.ERROR_TOO_MANY_OPEN_FILES;
			break;
		}

		return Promise.reject(
			new StorageException(e.message, code)
		);
	}

	_buildFullPath(basePath, blobName) {
		return basePath + "/" + blobName;
	}

	createPath(path) {
		return fsp.mkdir(path, {
			recursive : true,
			mode      : this._permission
		}).catch(e => this._errorHandler(e));
	}

	readBlob(basePath, blobName) {
		let path = this._buildFullPath(basePath, blobName);
		return fsp.readFile(path, {
			encoding : this._encoding
		}).catch(e => this._errorHandler(e));
	}

	writeBlob(basePath, blobName, data) {
		let path = this._buildFullPath(basePath, blobName);
		return fsp.writeFile(path, data, {
			encoding : this._encoding,
			mode     : this._permission
		}).catch(e => this._errorHandler(e));
	}

	appendBlob(basePath, blobName, data) {
		let path = this._buildFullPath(basePath, blobName);
		return fsp.appendFile(path, data, {
			encoding : this._encoding,
			mode     : this._permission
		}).catch(e => this._errorHandler(e));
	}

	deleteBlob(basePath, blobName) {
		let path = this._buildFullPath(basePath, blobName);
		return fsp.unlink(path).catch(e => this._errorHandler(e));
	}

	list(basePath) {
		return fsp.readdir(basePath, {
			encoding : this._encoding
		}).catch(e => this._errorHandler(e));
	}
}

module.exports = FileStorageConnector;