const SERVER_HOST = ''
const SERVER_PORT = ''

const HOTSPOT_DEFAULT_SSID = ''
const HOTSPOT_DEFAULT_PASS = ''

const beacon_valid_period = 10000	// How long util we report a missing beacon
const state_sync_period = 60000		// Sync with Web API

// Utilities
const fs = require('fs')
const path = require('path')
const http = require('http')
const crypto = require('crypto')
const EventEmitter = require('events')
const store = require('data-store')({ path: process.cwd() + '/store.json' })

// WebAPI master key
const smb_master_key = fs.readFileSync('./smb_master', 'utf8').trim()

// Persistent storage
//var store = new Storage('./storage')

// Interface for debug mode
const terminal = require(path.join(__dirname, 'terminal.js'))

// Bluetooth and internet gateway
const Net = require(path.join(__dirname, 'net.js'))

// Phone module driver
// This module is now obsolete, using the 'GPS' module instead
//const Sim = require(path.join(__dirname, 'sim7600.js'))
const GPS = require(path.join(__dirname, 'gps.js'))

// Battery charge monitor
const Battery = require(path.join(__dirname, 'battery-i2c.js'))

// Global state class
var state = new class extends EventEmitter {

	constructor () {
		super()

		this._timer = setInterval(this.tick.bind(this), 1000)
		this._beacon_timers = {}

		this._sync_timer = setInterval(this.sync.bind(this), state_sync_period)

		// Load all persistent data
		let beacons = store.get('beacons')
		if (typeof beacons == 'undefined') {
			beacons = { }
		} else {
			console.log('loaded beacons:', beacons)
		}

		let client_key = store.get('client_key')
		if (typeof client_key == 'undefined') {
			client_key = ''
		} else {
			console.log('loaded client key:', client_key)
		}

		let hotspot_ssid = store.get('hotspot_ssid')
		if (typeof hotspot_ssid == 'undefined') {
			hotspot_ssid = HOTSPOT_DEFAULT_SSID		// Fallback to default
		} else {
			console.log('loaded hotspot ssid:', hotspot_ssid)
		}

		let hotspot_pass = store.get('hotspot_pass')
		if (typeof hotspot_pass == 'undefined') {
			hotspot_pass = HOTSPOT_DEFAULT_PASS		// Fallback to default
		} else {
			console.log('loaded hotspot pass:', hotspot_pass)
		}

		let gps_location = store.get('gps_location')
		if (typeof gps_location == 'undefined') {
			gps_location = {		// Fallback to default
				a: null,			// latitude
				o: null,			// longitude
				t: Date.now(),		// timestamp (in bag time)
				s: 0				// location source: 0 - gps, 1 - gsm
			}
		} else {
			gps_location = JSON.parse(gps_location)
			console.log('loaded gps location:', gps_location)
		}

		this.sending = false
		this.state_cache = store.get('state_cache')
		if (typeof this.state_cache == 'undefined') {
			this.state_cache = []
		} else {
			console.log('loaded state cache, pending:', this.state_cache.length)
		}

		this._state = {
			w: {
				k: client_key		// WebAPI client key
			},
			s: {					// sim
				s: 0,				// connection: 0 - not connected, 1 - stabilizing, 2 - stable
				r: 1,				// lock: 0 - unlocked, 1 - pin locked, 2 - puk locked, 3 - permanently locked
				p: '',				// sim pin
				u: '',				// sim puk
			},
			h: {
				e: 0,				// hostpot enabled
				s: hotspot_ssid,	// currently used ssid
				p: hotspot_pass		// currently used password
			},
			l: gps_location,
			/* {
				a: 0,				// latitude
				o: 0,				// longitude
				t: Date.now(),		// timestamp (in bag time)
				s: 0				// location source: 0 - gps, 1 - gsm
			}*/
			v: {					// battery
				p: 0				// percentage
			},
			b: beacons				// ble beacons
			/*
			b: {
				'mac_address': {
					a: '',			// bluetooth mac address
					n: '',			// name
					l: 0,			// time last found in proximity
					p: 0,			// paired or not
					t: 0,			// type code
					e: false		// essential
				}
			}
			*/
		}

		console.log('INITIAL STATE', this._state)

		// Terminal initial
		terminal.hotspot = {
			on: this._state.h.e,
			ssid: this._state.h.s,
			pass: this._state.h.p
		}
		terminal.beacons = this._state.b
	}

	// Report that the state has changed
	_mutation (path = '') {

		switch (path) {
			case '.b': // Save beacons
				store.set('beacons', this._state.b)
				store.save()
				break

			case '.w.k': // Save client key
				store.set('client_key', this._state.w.k)
				store.save()
				break

			case '.h': // Save hotspot
				store.set('hotspot_ssid', this._state.h.s)
				store.set('hotspot_pass', this._state.h.p)
				store.save()
				break

			case '.l': // Save location
				store.set('gps_location', JSON.stringify(this._state.l))
				store.save()
				break
		}
	}

	get raw () { return this._state }

	set sim_stable (v) { this._state.s.s = v ? 1 : 0; this._mutation() }
	get sim_stable () { return this._state.s.s == 1 ? true : false }

	set sim_pinok (v) { this._state.s.r = v ? 1 : 0; this._mutation() }
	get sim_pinok () { return this._state.s.r == 1 ? true : false }

	set sim_pin (v) { this._state.s.p = v; this._mutation() }
	get sim_pin () { return this._state.s.p }

	set sim_puk (v) { this._state.s.u = v; this._mutation() }
	get sim_puk () { return this._state.s.u }

	// Location

	setLocation (lat, lon, source = 0) {

		let timestamp = Date.now()

		this._state.l.a = lat
		this._state.l.o = lon
		this._state.l.t = timestamp
		this._state.l.s = source

		// Update
		net.send({
			r: 13,
			a: lat,
			o: lon,
			t: timestamp,
			s: source
		})

		// Notify
		this._mutation('.l')
	}

	// Battery

	setBatteryCharge (percentage) {

		if (percentage != this._state.v.p) {
			this._state.v.p = percentage

			// Update
			net.send({
				r: 12,
				p: percentage
			})
		}

	}

	// Hotspot

	setHotspot (enabled, ssid, password) {

		if (enabled != this._state.h.e) {
			this._state.h.e = enabled
			this._state.h.s = ssid
			this._state.h.p = password

			// Update
			net.send({
				r: 16,
				e: enabled,
				s: ssid,
				p: password
			})
		}

		// Notify
		this._mutation('.h')
		terminal.hotspot = {
			on: enabled,
			ssid: ssid,
			pass: password
		}
	}

	// Beacons

	setBeacon (address, entry) {

		// Update the beacon
		if (this._state.b.hasOwnProperty(address)) {
			// Update
			this._state.b[address].c = entry.c		// Color should not change, but there it is
			this._state.b[address].d = entry.d		// Update distance
			this._state.b[address].l = Date.now()	// Update timestamp
		} else {
			// Insert
			this._state.b[address] = entry			// Color, distance and name
			this._state.b[address].l = Date.now()
			this._state.b[address].p = 0
		}

		// Notify
		this._mutation('.b')
		terminal.beacons = this._state.b
		this._queue_beacon_update(address)
	}

	linkBeacon (address, name = '', type = 0, essential = false) {
		console.log('linking beacon', address)

		if (this._state.b.hasOwnProperty(address)) {
			this._state.b[address].p = 1
			this._state.b[address].n = name
			this._state.b[address].t = type
			this._state.b[address].e = essential

			// Notify
			this._mutation('.b')
			terminal.beacons = this._state.b
			return 1
		} else {
			return 0
		}
	}

	unlinkBeacon (address) {
		console.log('unlinking beacon', address)

		if (this._state.b.hasOwnProperty(address)) {
			this._state.b[address].p = 0

			// Notify
			this._mutation('.b')
			terminal.beacons = this._state.b
			return 1
		} else {
			return 0
		}
	}

	updateBeacon (address, name = '', type = 0, essential = false) {
		console.log('updating beacon', address)

		if (this._state.b.hasOwnProperty(address)) {
			this._state.b[address].n = name
			this._state.b[address].t = type
			this._state.b[address].e = essential

			// Notify
			this._mutation('.b')
			terminal.beacons = this._state.b
			return 1
		} else {
			return 0
		}
	}

	_queue_beacon_update (address) {
		if (this._beacon_timers.hasOwnProperty(address)) {
			clearTimeout(this._beacon_timers[address])
		}
		this._beacon_timers[address] = setTimeout(function () {
			delete this.that._beacon_timers[this.address]

			net.send({
				r: 3,
				k: this.address,
				v: this.that._state.b[this.address]
			})
		}.bind({ that: this, address }), 5000)
	}

	tick () {
		// Invalidate beacons
		let time = Date.now()
		Object.keys(this._state.b).forEach((address) => {
			if (this._state.b[address].l < (time - beacon_valid_period)) {
				if (this._state.b[address].p == 0) {
					// This beacon is not linked, so drop it
					delete this._state.b[address]

					// Notify
					this._mutation('.b')
					terminal.beacons = this._state.b

					net.send({
						r: 4,
						k: address
					})
				} else {
					if (this._state.b[address].d !== 2) {
						// This beacon is linked, mark it as lost
						this._state.b[address].d = 2

						// Notify
						this._mutation('.b')
						terminal.beacons = this._state.b

						this._queue_beacon_update(address)
					}
				}
			}
		})
	}

	// WebAPI
	resetClientKey () {

		//var token = crypto.randomBytes(32).toString('base64')

		// Generate a simple password
	    let charset = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
	    let token = ''
	    for (let i = 0, n = charset.length; i < 8; ++i) {
	        token += charset.charAt(Math.floor(Math.random() * n))
	    }

		var data = JSON.stringify({
			data: {
				token,
				type: 'client'
			}
		})

		console.log('Resetting client key')

		return new Promise((resolve, reject) => {

			var post_req = http.request({
				host: SERVER_HOST,
				port: SERVER_PORT,
				path: '/data/' + net.device_name + '/token/reset',
				method: 'PUT',
				headers: {
					'Content-Type': 'application/json',
					'Content-Length': Buffer.byteLength(data, 'utf8'),
					'x-smb-auth-token': smb_master_key
				}
			}, (res) => {
				res.setEncoding('utf8')
				let result = ''
				res.on('data', (chunk) => {
					result += chunk
				})
				res.on('end', async () => {

					// Process results
					var response = null,
						valid = true

					try {
						response = JSON.parse(result)
						console.log('KEY RESET RESPONSE', response)

						// If successful
						if (response.ok) {

							this._state.w.k = token
							this._mutation('.w.k')
							console.log('CLIENT TOKEN UPDATED: ', token)

							resolve(token)
							return
						}

						// Something went wrong
						console.log('Client token could not be saved due to an unexpected error' + response)
						reject()

					} catch (err) {
						// Something went wrong
						console.log('Client token could not be saved' + err.message)
						reject()
					}
				})
			})

			post_req.on('error', (e) => {
				// There was an error during the upload, request failed
				console.log('Client token could not be saved' + e.message)
				reject()
			})

			post_req.write(data)
			post_req.end()
		})

	}

	// Sending state data to server
	async sync () {

		// Happens every [state_sync_period] milliseconds

		console.log(this._state.b);

		// Create a "state stamp"
		let state_stamp = {
			ts: Date.now(),
			lon: this._state.l.o,
			lat: this._state.l.a,
			battery: this._state.v.p,
			phone_active: (net.ble_connected && net.ble_authenticated) || net.ppp_connected,
			missing_gadgets: Object.values(this._state.b).reduce((accumulator, value) => { return accumulator + ( (value.e && (value.d == 2) ) ? 1 : 0 ) }, 0),
			gadgets: Object.values(this._state.b).reduce((accumulator, value) => { return accumulator + ( (value.e && (value.d < 2) ) ? 1 : 0 ) }, 0)
		}

		console.log(state_stamp);

		// Add a state stamp to the cache, and save it
		this.state_cache.unshift(state_stamp)
		store.set('state_cache', this.state_cache)
		store.save()
		console.log('New STATE STAMP, now pending: ', this.state_cache.length)

		// Send ---------------------------------------------------
		if (this.sending) return
		this.sending = true

		// Iterate through all the cached states
		while (this.state_cache.length > 0) {

			// Pick the oldest of the states
			let stamp_to_send = this.state_cache[this.state_cache.length - 1]

			// Try to send it
			try {
				// Make the http request
				await this.post(stamp_to_send)

				// Remove the last stamp that we just sent, and save the cache
				this.state_cache.pop()
				store.set('state_cache', this.state_cache)
				store.save()

				console.log('STATE STAMP synced, now pending: ', this.state_cache.length)
				await new Promise((resolve) => setTimeout(resolve, 1000))

			} catch (error) {

				console.log(error)

				// We could not send the stamp, let's wait a minute and let it try again
				await new Promise((resolve) => setTimeout(resolve, 60000))
			}
		}

		this.sending = false
	}

	post (state_stamp) {

		var data = JSON.stringify({
			data: state_stamp
		})

		return new Promise((resolve, reject) => {

			var post_req = http.request({
				host: SERVER_HOST,
				port: SERVER_PORT,
				path: '/data/' + net.device_name + "/data",
				method: 'POST',
				headers: {
					'Content-Type': 'application/json',
					'Content-Length': Buffer.byteLength(data, 'utf8'),
					'x-smb-auth-token': smb_master_key
				}
			}, (res) => {
				res.setEncoding('utf8')
				let result = ''
				res.on('data', (chunk) => {
					result += chunk
				})
				res.on('end', async () => {

					// Process results
					var response = null,
						valid = true

					try {
						response = JSON.parse(result)
						console.log('SYNC RESPONSE', response)

						// If successful
						if (response.ok) {
							resolve()
							return
						}

						// Something went wrong
						reject()

					} catch (err) {
						// Something went wrong
						console.log('SYNC ERROR',err)
						reject()
					}
				})
			})

			post_req.on('error', (e) => {
				// There was an error during the upload, request failed
				reject()
			})

			post_req.write(data)
			post_req.end()
		})
	}
} ()

// Create the network interface
var net = new Net(state, terminal)

// Create the sim interface
var gps = new GPS(state, terminal)

// Create the battery charge monitor instance
var battery = new Battery(state, terminal)

// Check if the hotspot is active or not
net.hotspotCheck()
