/* The Net class handles messaging between the SMB and a single connected device. Communication can
 * happen over bluetooth low energy (BLE), internet, or other means. The connection state of the
 * device and presence of registered BLE beacons is also monitored through this class.
 */

// Dependencies
const fs = require('fs')
const md5 = require('blueimp-md5')
const EventEmitter = require('events')
const { exec, execSync } = require('child_process')
const WebSocket = require('ws')

const WEBSOCKET_URL = '[REDACTED]' // expected format: "ws://[host]:[port]"

// This is the Raspberry Pi's processor serial number, it is expected to be an 8 octet hex string
const device_serial = fs.readFileSync('/proc/cpuinfo', 'utf8').match(/^Serial[^:]*\:\ (.*)$/m)[1]
// The device name contains the uppercase version of the last 4 octets of the Pi's serial number,
// as it appears that the more significant hex digits are not used
const device_name = 'SMB-' + device_serial.substring(8, 16).toUpperCase()
const device_pin = fs.readFileSync('./smb_pin', 'utf8').trim()

// Bluetooth configuration and initialization ------------------------------------------------------

/* Find the hci device ids for the two devices
 * Here, we assume that there are two hci devices. One integrated into the pi, and an usb dongle
 * acting as secondary. The below code determines which is which, by looking at their bus.
 */
var hci_devices = [ { bus: 'UNDETECTED', address: '' }, { bus: 'UNDETECTED', address: '' } ]
try {
	const hci_regex = /hci(.):.*Bus:\ (.*)\n.*BD Address: ((?:[0-9A-F]{2}\:){5}[0-9A-F]{2})/g
	let result = execSync('hciconfig', { encoding: 'utf8' })
	let hci = null
	while ((hci = hci_regex.exec(result)) !== null) {
		let [ , id, bus, address ] = hci
		hci_devices[id] = { bus, address }
	}
} catch (err) {
	// We failed somewhere apparently
	console.err(err)
} finally {
	// Set the peripheral mode device
	let bleno_device = hci_devices.findIndex(hci => hci.bus == 'UART')
	process.env['BLENO_HCI_DEVICE_ID'] = (bleno_device >= 0) ? bleno_device : 0

	// Set the central mode device
	let noble_device = hci_devices.findIndex(hci => hci.bus == 'USB')
	process.env['NOBLE_HCI_DEVICE_ID'] = (noble_device >= 0) ? noble_device : 1
}

process.env['BLENO_ADVERTISING_INTERVAL'] = 500

console.log('BLENO: ' + process.env['BLENO_HCI_DEVICE_ID'])
console.log('NOBLE: ' + process.env['NOBLE_HCI_DEVICE_ID'])

// Now begin with the drivers
const noble = require('noble')
const bleno = require('bleno')

// The service and characteristic uuids are the same across all SMBs. Uuids of this length are not
// subjet to special licensing, and they were generated by authority sanctioned tools.
// They are used to identify the device as an SMB over bluetooth.
const service_uuid = '[REDACTED]'
const characteristic_uuid = '[REDACTED]'

// The service uuids for the beacons are corresponding to their color to aid in identifying them.
const beacon_uuid = '[REDACTED]'
const beacon_colors = [ 'unspecified', 'black', 'white', 'cyan', 'blue', 'green', 'red' ]
// The duration and frequency of beacon scans. Just to be on the safe side, the interval should be
// at least an order of magnitude larger than the duration.
const beacon_track_duration = 8000
const beacon_track_interval = 10000

// BLE packet constraints
const packet_size = 512			// Maximum size of a packet in bytes
const packet_header_size = 6	// A packet starts with a header this size, in bytes
const packet_data_size = packet_size - packet_header_size

// PPP
const ppp_regex = /ppp(.).*\n.*inet ([^\ ]*)/g
const ppp_seed_1 = '[REDACTED]'
const ppp_seed_2 = '[REDACTED]'
if (hci_devices.hasOwnProperty(process.env['BLENO_HCI_DEVICE_ID'])) {
	var ppp_channel_id = (device_serial.substring(8, 16) + hci_devices[process.env['BLENO_HCI_DEVICE_ID']].address.split(':').join('') + md5(ppp_seed_2 + md5(ppp_seed_1 + device_pin)).substring(0, 16)).toUpperCase()
} else {
	console.log('No primary hci device')
	process.exit()
}
const ppp_smb_id = ppp_channel_id + '-1'
const ppp_app_id = ppp_channel_id + '-2'

module.exports = class Net extends EventEmitter {

	constructor (state, terminal) {
		super()

		console.log('Starting SMB net...')
		console.log('Channel:' + ppp_smb_id)

		this.device_name = device_name

		this.state = state	// Save a reference to the global state object
		this.terminal = terminal
		this.terminal.processor_id = device_serial.substring(8, 16).toUpperCase()
		this.terminal.hci_peripheral = `hci${process.env['BLENO_HCI_DEVICE_ID']}  ${(hci_devices[process.env['BLENO_HCI_DEVICE_ID']].bus.padEnd(4))}  ${(hci_devices[process.env['BLENO_HCI_DEVICE_ID']].address)}`
		this.terminal.hci_central = `hci${process.env['NOBLE_HCI_DEVICE_ID']}  ${(hci_devices[process.env['NOBLE_HCI_DEVICE_ID']].bus.padEnd(4))}  ${(hci_devices[process.env['NOBLE_HCI_DEVICE_ID']].address)}`

		// Define Bluetooth peripheral mode behaviour ----------------------------------------------

		this.ble_powered = false
		this.ble_connected = false
		this.ble_challenge = ''
		this.ble_authenticated = false
		this.ble_ps_count = 0	// Number of packets sent
		this.ble_pr_count = 0	// Number of packets received
		this.ble_queue = []
		this.ble_processing = false
		this.ble_send_payload = null
		this.ble_send_cursor = 0
		this.ble_send_resolve = null
		this.ble_send_reject = null
		this.ble_send_notify = null

		/*
		 * The bluetooth low energy mode accepts a single connection only. Once we connect to a
		 * device, all packets that are exchanged will be addressed to that device, and no other
		 * device can initiate connection.
		 */
		bleno.setServices([
			new bleno.PrimaryService({
				uuid: service_uuid,
				characteristics: [
					new SMBCharacteristic(this)
				]
			})
		])

		/* The bleno states are the following:
		 * unknown | resetting | unsupported | unauthorized | poweredOff | poweredOn
		 * Of these states, only the 'poweredOn' state means that the bluetooth interface is ready
		 * and can be used to create a connection.
		 */
		bleno.on('stateChange', (state) => {
			if (state == 'poweredOn') {
				this.ble_powered = true
				bleno.startAdvertising(device_name, [ service_uuid ])
			} else {
				this.ble_powered = false
				bleno.stopAdvertising()
			}
		})

		/* A connection can only be initiated from the remote device (phone app). For it to be able
		 * to detect the SMB, it has to advertise itself as a connectable ble peripheral. Once a
		 * connection is made, the advertising will stop, preventing another device to connect.
		 * (This behaviour is defined in the BLE standard, and cannot be changed in bleno.)
		 */
		bleno.on('advertisingStart', function(error) {
			// Advertising started
			console.log('BLENO - advertising started: ' + (error ? 'error ' + error : 'success'))
		})

		bleno.on('advertisingStop', function() {
			// Advertising stopped
		})

		bleno.on('accept', (clientAddress) => {
			// A connection has been made
			this.ble_connected = true
			this.terminal.ble_conn = true
			console.log('BLENO - connect: ' + clientAddress)
		})

		bleno.on('disconnect', (clientAddress) => {
			// The connection was lost
			this.ble_connected = false
			this.terminal.ble_conn = false
			console.log('BLENO - disconnect: ' + clientAddress)
		})

		// Define Bluetooth central mode behaviour -------------------------------------------------

		this.noble_powered = false
		this.noble_period_timer = 0
		this.noble_duration_timer = 0

		/* Secondary hci device for scanning beacon signals. We subscribe to power state changes
		 * below, and start/stop the periodic scan accordingly.
		 */
		noble.on('stateChange', (state) => {
			console.log('NOBLE - State: ' + state)
			if (state === 'poweredOn') {
				this.noble_powered = true
				this.enableBeaconTracking()
			} else {
				this.noble_powered = false
				this.disableBeaconTracking()
			}
		})

		noble.on('scanStart', () => { this.terminal.hci_scanning = true })
		noble.on('scanStop', () => { this.terminal.hci_scanning = false })

		noble.on('discover', (peripheral) => {

			try {
				let data = peripheral.advertisement.manufacturerData

				// If the manufacturer data is missing, ignore this one
				if (typeof data === 'undefined') {
					return
				}

				// If the manufacturer data is too short for our expected ibeacon frame, ignore
				if (data.length < 25) {
					return
				}

				let uuid = data.slice(4, 20).toString('hex')

				// If the extracted uuid does not correspond to an SMB beacon, ignore it
				if (uuid != beacon_uuid) {
					return
				}

				let major = data.slice(20, 22).readUInt16BE(0), // Used for color id in our system
					minor = data.slice(22, 24).readUInt16BE(0), // Open field, unused at the moment
					baseline_rssi = data.slice(24, 25).readInt8(0),
					relative_rssi = (baseline_rssi - peripheral.rssi)

				// Send this information to the global state keeper
				this.state.setBeacon(peripheral.address, {
					c: data.slice(20, 22).readUInt16BE(0),
					d: (relative_rssi < 0 ? 0 : 1)
				})

			} catch (err) {

				// Some error occurred during the interpretation of the data, we ignore these
				console.log(err)
			}
		})

		// Define PPP connection behaviour ---------------------------------------------------------

		this.ppp_connected = false
		this.ppp_client = new PPP_Client(ppp_smb_id, ppp_app_id)
		this.ppp_ps_count = 0

		this.ppp_client.on('state-change', (state) => {
			this.ppp_connected = state
			this.terminal.tunnel_conn = state
			console.log(this.ppp_connected ? 'CONNECTED TO REMOTE DEVICE' : 'DISCONNECTED FROM REMOTE DEVICE')
		})

		this.ppp_client.on('receive', (payload) => {
			this.receive(payload)
		})

		// Disable hotspot on startup
		this.disableHotspot()
	}

	// API

	send (data, force_ble = false) {
		if (this.ppp_connected && !force_ble) {
			// Send over PPP if connected
			console.log('sending mesage over TUNNEL', data)
			this.ppp_client.send_message(data)
		} else if (this.ble_connected) {
			// Send over BLE if connected
			console.log('sending mesage over BLE', data)
			this.ble_queue.push(BlePayload.Outgoing(data))
			this.ble_process()
		} else {
			// Not connected
			console.log('not sending', data)
		}
	}

	// Internals

	// PPP specifics -------------------------------
	/*
	ppp_check () {
		exec('ifconfig', (err, stdout, stderr) => {
			if (err) {
				console.err(err)
				// The command could not be executed
				return
			}

			// Get a list of all the ppp network interfaces
			let ppp_devices = []
			let ppp = null
			while ((ppp = ppp_regex.exec(stdout)) !== null) {
				let [ , id, address ] = ppp
				ppp_devices[id] = { address }
			}

			// TODO: Update ppp status based on our findings
			// ...
		})
	}
	*/

	ppp_send (payload) {
		this.ppp_client.send(payload)
	}

	// BLE peripheral specifics --------------------

	// Process the queue: send payloads
	async ble_process () {

		// Make sure this only happens in this thread
		if (this.ble_processing) return
		this.ble_processing = true

		// Handle all pending transmissions
		while (this.ble_queue.length > 0) {

			let payload = this.ble_queue.shift()
			console.log('sending payload')
			try {
				await this.ble_send(payload)
			} catch (error) {
				console.log('sending failed: ', error)
			}
		}

		// Release
		this.ble_processing = false
	}

	ble_send (payload) {

		// Set the payload to be sent
		this.ble_send_payload = payload
		this.ble_send_cursor = 0

		// Promise that will be resolved when the payload has been delivered
		let promise = new Promise((resolve, reject) => {
			this.ble_send_resolve = resolve
			this.ble_send_reject = reject
		})

		// Notify the remote device that we have a payload to send
		if (this.ble_send_notify !== null) {
			this.ble_send_notify(Buffer.from('r,' + payload.length))
		} else {
			throw 'not subscribed'
		}

		return promise
	}

	receive (request) {
		console.log('Message received: ', request)

		// Ignore anything that is not an object
		if (typeof request !== 'object') return

		// Ignore any requests without a type identifier
		if (!request.hasOwnProperty('r')) return

		// Let's see what type of request we are dealing with, and handle accordingly
		switch (request.r) {

			/* Request ids:
			 * 1 - ping < >
			 * 2 - sync < >
			 * 3 - update beacon >
			 * 4 - delete beacon >
			 * 5 - link beacon < >
			 * 6 - unlink beacon < >
			 * 7 - set beacon < >
			 *
			 * 10 - ble auth challenge < >
			 * 11 - ble auth response < >
			 * 14 - set hotspot < >
			 * 15 - renew client key < >
			 * 16 - hotsport report >
			 */

			case 10: // auth challenge
				// Reply with the session challenge
				this.send({
					i: request.i,
					r: 10,
					c: this.auth_challenge(),
					t: Date.now()
				}, true)
				break

			case 11: // auth response
				// Reply with the result of the authentication attempt
				this.send({
					i: request.i,
					r: 10,
					s: this.auth_response(request.a) ? 1 : 0
				}, true)
				break

			case 1: // ping
				// Reply with pong
				this.send({
					i: request.i,
					r: 1
				})
				break

			case 2: // sync
				this.send({
					i: request.i,
					r: 2,
					s: this.state.raw
				})
				break

			case 5: // link beacon
				this.send({
					i: request.i,
					r: 5,
					s: this.state.linkBeacon(request.a, request.n, request.t, request.e)
				})
				break

			case 6: // unlink beacon
				this.send({
					i: request.i,
					r: 6,
					s: this.state.unlinkBeacon(request.a)
				})
				break

			case 7: // update beacon
				this.send({
					i: request.i,
					r: 7,
					s: this.state.updateBeacon(request.a, request.n, request.t, request.e)
				})
				break

			case 14: // set hotspot

				// Set ssid and password
				(async () => {
					await this.setHotspotConfig((request.e == 1), request.s, request.p)
					if (request.e == 1) {
						// Wait a second for hostapd to restart
						await new Promise(resolve => setTimeout(resolve, 1000))
					}

					// Start / stop wi-fi
					let success = ((request.e == 1) ? await this.enableHotspot() : await this.disableHotspot())

					this.send({
						i: request.i,
						r: 14,
						s: (success ? 1 : 0)
					})
				})()
				break

			case 15: // reset client key

				(async () => {
					// Attempt to reset the client key
					try {
						var key = await this.state.resetClientKey()
					} catch (e) {
						// Fail
						this.send({
							i: request.i,
							r: 15,
							s: 0
						})
						return
					}

					// Success
					this.send({
						i: request.i,
						r: 15,
						s: 1,
						k: key
					})
				})()
				break

			default: // Any unimplemented requests..
				this.send(request)	// Just echo the request as it is
				break
		}
	}

	// Generate a challenge string, which will be sent to an SMB trying to authenticate
	auth_challenge (length = 10) {
		let result = ''
		let characters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789'
		let charactersLength = characters.length
		for (let i = 0; i < length; i++) {
			result += characters.charAt(Math.floor(Math.random() * charactersLength))
		}

		this.ble_challenge = result
		return result
	}

	// Generate a hash from the challenge string and the SMB pin
	auth_response (response) {

		let expected_response = md5(this.ble_challenge + device_pin)

		if (expected_response == response) {
			// Authentication successful
			this.ble_authenticated = true
			console.log('AUTH SUCCESS')
			return true
		} else {
			// Authentication failed, reject connection
			this.ble_authenticated = false
			console.log('AUTH FAIL')
			return false
		}
	}

	// BLE central specifics -----------------------

	enableBeaconTracking () {
		// Reset the timer
		if (this.noble_period_timer !== 0) clearInterval(this.noble_period_timer)
		this.noble_period_timer = setInterval(this.trackBeacons, beacon_track_interval)
		// Start scanning immediately
		this.trackBeacons()
	}

	disableBeaconTracking() {
		// Clear timer
		if (this.noble_period_timer !== 0) clearInterval(this.noble_period_timer)
		// Stop scanning
		noble.stopScanning()
	}

	trackBeacons () {
		// Start the scan.. (scan discovery handling is defined in the class constructor)
		noble.startScanning([], true)

		this.noble_duration_timer = setTimeout(() => {
			this.noble_duration_timer = 0

			// Stop the scan after the duration has ended
			noble.stopScanning()

		}, beacon_track_duration)
	}

	// Hotspot

	async hotspotCheck () {

		// Check to see if the wi-fi is on
		let stdout = ''
		try {
			stdout = await new Promise((resolve, reject) => {
				exec('ifconfig', (err, stdout, stderr) => {
					if (err) {
						// There was an error
						reject(err)
						return
					}

					// No error
					resolve(stdout)
				})
			})
		} catch (err) {
			console.err(err)
			// The command could not be executed
			return
		}

		// Get a list of all the wlan network interfaces
		const wlan_regex = /wlan(.).*\n.*inet.{0,1} ([^\ ]*)/g

		let wlan_devices = []
		let wlan = null
		while ((wlan = wlan_regex.exec(stdout)) !== null) {
			let [ , id, address ] = wlan
			wlan_devices[id] = { address }
		}

		// Check the hostapd config
		try {
			var config = await new Promise((resolve, reject) => {
				fs.readFile('/etc/hostapd/hostapd.conf', { encoding: 'utf8' }, (err, data) => {
	  				if (err) {
						// There was an error
						reject(err)
						return
					}

					// No error
	  				resolve(data)
				})
			})
		} catch (err) {
			console.err(err)
			// The command could not be executed
			return
		}

		// Parse the config file to read the ssid and password
		let ssid = ''
		let password = ''
		try {
			const wlan_config_ssid_regex = /^ssid=(.*)$/gm
			const wlan_config_password_regex = /^wpa_passphrase=(.*)$/gm

			let res = ''
			if ((res = wlan_config_ssid_regex.exec(config)) !== null) {
				console.log(res)
				ssid = res[1]
			}
			if ((res = wlan_config_password_regex.exec(config)) !== null) {
				console.log(res)
				password = res[1]
			}
		} catch (err) {
			console.err(err)
			// The command could not be executed
			return
		}

		// Update hotspot status based on our findings
		if (wlan_devices.hasOwnProperty(0)) {
			// On
			console.log(`hotspot is on (ssid:${ssid}; pass:${password})`)
			this.state.setHotspot(1, ssid, password)
		} else {
			// Off
			console.log(`hotspot is off (ssid:${ssid}; pass:${password})`)
			this.state.setHotspot(0, ssid, password)
		}
	}

	async enableHotspot() {
		console.log('HOTSPOT turning on')

		try {
			let stdout = execSync('ifconfig wlan0 up')
		} catch (err) {
			console.err(err)
			// The command could not be executed
			return false
		}

		await new Promise(resolve => setTimeout(resolve, 3000))
		this.hotspotCheck()
		return true
	}

	async disableHotspot () {
		console.log('HOTSPOT turning off')

		try {
			let stdout = execSync('ifconfig wlan0 down')
		} catch (err) {
			console.err(err)
			// The command could not be executed
			return false
		}

		await new Promise(resolve => setTimeout(resolve, 3000))
		this.hotspotCheck()
		return true
	}

	async setHotspotConfig (enabled, ssid, password) {

		const config_content =
`interface=wlan0
driver=nl80211
ssid=${ssid}
hw_mode=g
channel=7
wmm_enabled=0
macaddr_acl=0
auth_algs=1
ignore_broadcast_ssid=0
wpa=2
wpa_passphrase=${password}
wpa_key_mgmt=WPA-PSK
wpa_pairwise=TKIP
rsn_pairwise=CCMP
`
		// Overwrite hostapd config
		try {
			await new Promise((resolve, reject) => {
				fs.writeFile('/etc/hostapd/hostapd.conf', config_content, { encoding: 'utf8' }, (err) => {
					if (err) {
						// There was an error
						reject(err)
						return
					}

					// No error
					resolve()
				})
			})
		} catch (err) {
			console.err(err)
			// The command could not be executed
			return
		}

		// Restart hostapd
		if (enabled) {
			let stdout = ''
			try {
				stdout = await new Promise((resolve, reject) => {
					exec('systemctl restart hostapd', (err, stdout, stderr) => {
						if (err) {
							// There was an error
							reject(err)
							return
						}

						// No error
						resolve(stdout)
					})
				})
			} catch (err) {
				console.err(err)
				// The command could not be executed
				return
			}
		}
	}
}

class SMBCharacteristic extends bleno.Characteristic {

	constructor (net) {
		super({
			uuid: characteristic_uuid,
			properties: ['read', 'write', 'notify'],
			value: null
		})

		// Save a reference to the Net class
		this.net = net

		// Read cache for incoming write requests
		this._cache = {}
	}

	// Sending data to remote device
	onReadRequest (offset, callback) {

		// Check to see if we have something to send
		if (this.net.ble_send_payload === null) {
			// If not, just send a zero, letting the app know we're good
			callback(this.RESULT_SUCCESS, new Buffer([0]))
			console.log('BLE SEND ERROR: nothing to send')
			return
		}

		// Check if we are within bounds
		if (this.net.ble_send_cursor >= this.net.ble_send_payload.length) {
			// Nope, zero it out
			callback(this.RESULT_SUCCESS, new Buffer([0]))
			console.log('BLE SEND ERROR: request is out of bounds')
			return
		}

		let chunk = this.net.ble_send_payload.chunk(this.net.ble_send_cursor)
		let end = Math.min(offset + 22, chunk.length)
		let piece = chunk.subarray(offset, end)

		console.log('BLE SEND: offset = ' + offset + ' piece = ' + piece.toString())
		callback(this.RESULT_SUCCESS, piece)

		// Check if the chunk has been completed
		if (end == chunk.length) {
			if (this.net.ble_send_cursor < this.net.ble_send_payload.length - 1) {
				// There are more chunks, so advance the cursor
				console.log('BLE SEND: chunk sent, advancing')
				this.net.ble_send_cursor++
			} else {
				// This payload has been sent completely
				console.log('BLE SEND: payload sent')
				this.net.ble_send_resolve()
			}
		}
	}

	// Receiving data from remote device
	onWriteRequest (buffer, offset, withoutResponse, callback) {

		// Header
		let header = {
			series_nr: buffer.readInt32BE(0),
			packet_nr: buffer.readUInt8(4),
			packet_count: buffer.readUInt8(5)
		}

		// Data
		// If the series number is not in the cache yet, let's create a new incoming payload for it
		if (!this._cache.hasOwnProperty('_' + header.series_nr)) {
			console.log('BLE RECEIVE: New payload: ', header.series_nr, header.packet_count)
			this._cache['_' + header.series_nr] = BlePayload.Incoming(header.packet_count)
		}
		// Add the packet data to the proper incoming payload
		console.log('BLE RECEIVE: Chunk for payload:', header.series_nr, header.packet_nr)
		this._cache['_' + header.series_nr].store(header.packet_nr - 1, buffer.subarray(packet_header_size))

		// Has this payload been completed just now?
		if (header.packet_nr == header.packet_count) {

			// Report that we have received a payload
			console.log('BLE RECEIVE: Payload complete')
			this.net.receive(this._cache['_' + header.series_nr].value)

			// Remove from the cache
			delete this._cache['_' + header.series_nr]
		}

		callback(this.RESULT_SUCCESS)
	}

	onSubscribe (maxValueSize, updateValueCallback) {
		console.log('EchoCharacteristic - onSubscribe, max length: ', maxValueSize)
		this.net.ble_send_notify = updateValueCallback
	}

	onUnsubscribe () {
		console.log('EchoCharacteristic - onUnsubscribe')
		this.net.ble_send_notify = null
	}

	/*
	notify () {
		console.log('Sending notification')
		if (this._updateValueCallback) {
			console.log('callback found')
			this._updateValueCallback(Buffer.from('send'))
		} else {
			console.log('callback not found')
		}
	}
	*/
}

/* A payload represents a single message sent one way over bluetooth. It will either be an incoming
 * or outgoing message. The payload may be split into packets of a specific max size, so it can be
 * transferred over BLE.
 */
class BlePayload {

	static Incoming (length) {
		let payload = new BlePayload('in')
		payload._length = length
		payload._series = new Array(payload._length)
		return payload
	}

	static Outgoing (value) {
		let payload = new BlePayload('out')
		payload.value = value
		return payload
	}

	constructor (mode) {
		this._mode = mode
		this._time = Date.now()
		this._length = null
		this._series = null			// Array of buffers that contain chunks of the payload
	}

	// Is this an incoming or an outgpoing payload?
	get mode () { return this._mode }

	// How many packets are expected in this series?
	get length () { return this._length }

	// Get a chunk of binary data
	chunk (i) { return this._series[i] }

	// Add (or replace) chunk data (incoming payloads)
	store (offset, buffer) {
		this._series[offset] = buffer
	}

	// Check whether or not all the chunks are present (incoming payloads)
	is_complete () {
		for (let i = 0; i < this._length; i++) {
			if (typeof this._series[i] === 'undefined')
				return false
		}
		return true
	}

	// Return the decoded payload object
	get value () {
		// Convert the series into the message object
		try {
			if (!this.is_complete())
				throw 'data incomplete'

			// Combine the strings encoded in the chunks
			let str = ''
			for (let i = 0; i < this._length; i++) {
				str += this._series[i].toString('utf8')
			}

			// Return the original json
			return JSON.parse(str)

		} catch (error) {
			throw 'invalid payload - ' + error
		}
	}

	// Set the payload object
	set value (value) {
		// Convert the message object into a series, and write it into a memory buffer
		let buffer = Buffer.from(JSON.stringify(value), 'utf8')

		// Slice the buffer to chunks no longer than packet_data_size bytes
		this._length = Math.ceil(buffer.length / packet_data_size)
		this._series = new Array(this._length)
		for (let i = 0; i < this._length; i++) {
			this._series[i] = buffer.subarray(i * packet_data_size, (i + 1) * packet_data_size)
		}
	}
}

class PPP_Client extends EventEmitter {

	constructor (me, partner) {
		super()

		console.log('Initializing socket connection')

		this._me = me
		this._partner = partner

		this._active = false
		this._connected = false
		this._socket = null
		this._seq = 0
		this._heartbeat_interval = null
		this._last_incoming = Date.now()

		//this._connection_timeout = null

		this.connect()
	}

	connect () {

		console.log('TUNNEL - Attempting to open socket')

		if (this._active) {
			console.log('Cannot open socket, because it is already active')
			return
		}

		// Recreate socket
		this._active = true
		this._socket = new WebSocket(WEBSOCKET_URL)

		/*
		// Handle Connection timeout
		this._connection_timeout = setTimeout(function () {
			// Clear timer
			clearTimeout(this._connection_timeout)
			this._connection_timeout = null

			// Check to see if the connection has been opened or not
			if (this._socket.readyState !== WebSocket.OPEN) {
				console.log('socket connection timeout')
				this.close()
			}
		}.bind(this), 5000)
		*/

		// Socket opened
		this._socket.on('open', e => {
			console.log("TUNNEL Socket Open")

			// Start heartbeats
			this._heartbeat_interval = setInterval(this.heartbeat.bind(this), 15000)

			this.send({
				type: "open",
				me: this._me,
				rec: this._partner
			})
		})

		this._socket.on('error', (e) => {
			console.log("TUNNEL Socket Error", e)
		})

		this._socket.on('close', (e) => {
			console.log("TUNNEL Socket Closed", e)

			// Stop heartbeats
			clearInterval(this._heartbeat_interval)
			this._heartbeat_interval = null

			this.onclose()
		})

		this._socket.on('message', (e) => {

			let payload = JSON.parse(e)

			this._last_incoming = Date.now()

			if (!payload.hasOwnProperty('type')) return
			switch (payload.type) {
				case 'ack':
				case 'control':
					//console.log('TUNNEL RECEIVE ACK/CNT')

					// We received a control or ack message, update the connection state
					let old_connection_state = this._connected
					this._connected = (payload.m.cli == 2)

					// If the state changed, let's notify
					if (this._connected !== old_connection_state) {
						this.emit('state-change', this._connected)
					}

					break
				case 'data':
					// We received a data message
					//console.log('TUNNEL RECEIVE', payload.d)
					this.emit('receive', payload.d)
					break
			}
		})
	}

	heartbeat () {

		// Connection timeout
		if (this._last_incoming < Date.now() - 30000) {
			this._socket.close()
			return
		}

		// Keepalive
		if (this._socket.readyState == 1) {
			//console.log('TUNNEL Heartbeat')
			this.send({ type: "ping" })
		}
	}

	send_message (data) {
		return this.send({
			type: "send",
			d: data,
		})
	}

	send (payload) {
		payload.seq = ++this._seq
		this._socket.send(JSON.stringify(payload))
		return payload.seq
	}

	close () {
		if (this._socket !== null) {
			// Close the socket
			this._socket.close()
		} else {
			// The socket was never open to begin with
			this.onclose()
		}
	}

	// This function will always be called when the SOCKET connection is lost / closed
	async onclose () {
		this._active = false

		let old_connection_state = this._connected
		this._connected = false

		// If the state changed, let's notify
		if (this._connected !== old_connection_state) {
			this.emit('state-change', this._connected)
		}

		/*
		// Clear timer
		clearTimeout(this._connection_timeout)
		this._connection_timeout = null
		*/

		// Wait a few seconds
		await new Promise((resolve) => { setTimeout(resolve, 5000) })

		// Reconnect
		this.connect()
	}
}
