const SerialPort = require('serialport')
const ReadLine = require('@serialport/parser-readline')
const path = require('path')
const Mutex = require('./mutex')
const channel = new Mutex()
const { exec, execSync } = require('child_process')

const PORT_ADDRESS = '/dev/gsm1'
const APN = ''
const SIM_PIN = '0000'
const SIM_PUK = '00000000'
const COMMAND_TIMEOUT = 10

const loc_interval = 60000 		// in ms
const gps_cycles = 5			// how many times we try to check for the gps location
const gps_cycle_time = 5000		// how long do we wait between tries
const gsm_cycles = 5			// how many times we try to check for the gsm location
const gsm_cycle_time = 1000		// how long do we wait between tries

module.exports = class Sim {

	constructor (state, terminal) {

		this.state = state	// Save a reference to the global state object
		this.terminal = terminal

		this.timer = null
		this.cache = []		// Input cache

		this.port = new SerialPort(PORT_ADDRESS, {
			autoOpen: false,
			echo: true,
			baudRate: 38500
		})

		this.port.on('error', (err) => {
			console.log('SIM - serial connection produced an error: ', err)
		})

		this.port.on('open', () => {
			console.log('SIM - serial connection is now open')
			this.terminal.cell_uart = 'OPEN'

			// Current command
			this.command = {
				command: '',
				resolve: null,
				reject: null
			}

			// Initiate connection
			this.state.sim_stable = false
			this.state.sim_pinok = false
			this.init()
		})

		this.port.on('close', () => {
			this.state.sim_stable = false
			this.state.sim_pinok = false
			console.log('SIM - serial connection is now closed')
			this.terminal.cell_uart = 'CLOSED'
		})

		// Split the input into lines
		this.port.parser = this.port.pipe(new ReadLine({ delimiter: String.fromCharCode(13) + String.fromCharCode(10) }))

		// Lines coming from the sim are written to the cache
		this.port.parser.on('data', (line) => {

			//console.log('SIM - input (' + line.replace('\r', '<CR>').replace('\n', '<LF>') + ')')

			// If the line is an 'OK', we process the already cached lines
			if (line == 'OK') {
				this.process()
				return
			}

			// If the line is an 'ERROR' we handle the cached lines as an error message
			if (line == 'ERROR') {
				this.error()
				return
			}

			// Otherwise just push the line into the cache
			this.cache.push(line)
		})

		this.port.open()
	}

	async init () {

		console.log('SIM - init')

		// Baud rate automatikus beállítása az AT parancs többszöri elküldésével
		while (!this.state.sim_stable) {
			try {
				await this.exec('AT')
				this.state.sim_stable = true
			} catch (err) {
				// Az AT parancs időtúllépéssel vagy hibával tért vissza, próba újra
				this.state.sim_stable = false
			}
		}

		console.log('SIM - stable')

		// Unlock sim
		await this.sim_unlock()

		console.log('SIM - ready')

		// Initiate the timer for recurring jobs
		setInterval(this.tick.bind(this), loc_interval)
	}

	// Recurring job
	async tick () {

		// Wait for channel access
		let mutex = await channel.request()

		// GPS coordinate query
		try {
			// Turn the gps on
			await this.exec('AT+CGNSPWR=1')

			// Try to get a gps fix, but stop trying after 20 secods
			for (let i = 0; i < gps_cycles; i++) {

				// Wait 1 sec
				await new Promise((resolve) => { setTimeout(resolve, gps_cycle_time) })

				let result = await this.exec('AT+CGNSINF')
				let pos_line = result.find(line => line.startsWith('+CGNSINF:'))
				if (typeof pos_line !== 'undefined') {
					let pos_data = pos_line.substring(10).split(',')
					if (pos_data[1] == '1') {

						// We have a gps fix!
						console.log('gps fix: ', pos_data[3], pos_data[4])

						// Send the location to the global state object
						this.state.setLocation(pos_data[3], pos_data[4], 0)
						break

					} else {

						// No fix, keep trying
						console.log('no gps fix')
					}
				} else {
					console.log('gps command fail')
				}
			}

			// Turn the gps off
			//await this.exec('AT+CGNSPWR=0')

		} catch (err) {

			console.log('SIM - ERROR:', err)

		}

		// GSM coordinate query

		/*
		await this.exec('AT+SAPBR=3,1,"Contype","GPRS"')			// Set connection type
		await this.exec('AT+SAPBR=3,1,"APN","internet.telekom"')	// Set APN
		await this.exec('AT+SAPBR=1,1')								// Open connection
		await this.exec('AT+SAPBR=2,1')								// Query IP address
		await this.exec('AT+CIPGSMLOC=1,1')							// Query GSM location
		await this.exec('AT+SAPBR=0,1')								// Close connection
		*/

		mutex.release()
	}

	// Jobs ----------------------------------------------------------------------------------------

	async sim_unlock () {

		// Wait for channel access
		let mutex = await channel.request()

		try {

			// Try to unlock the sim with the pin code, unless it's already unlocked
			let pin_tries = 0
			let puk_tries = 0
			while (!this.state.sim_pinok) {

				// Get the sim lock state
				let pin_state = await this.exec('AT+CPIN?')
				console.log(pin_state)

				switch (pin_state[1]) {

					// Sim is unlocked already
					case '+CPIN: READY':
						this.state.sim_pinok = true
						break

					// The sim is requesting a pin code
					case '+CPIN: SIM PIN':
						if (sim_pin == '') throw 'no pin was provided'
						if (pin_tries > 0) throw 'provided sim pin is incorrect'
						pin_tries++
						// Send pin
						await this.exec('AT+CPIN=' + SIM_PIN)
						// Wait 2 secs
						await new Promise(done => setTimeout(done, 2000))
						break

					// The sim is requesting a puk code
					case '+CPIN: SIM PUK':
						if (sim_pin == '') throw 'no pin was provided'
						if (sim_puk == '') throw 'no puk was provided'
						if (puk_tries > 0) throw 'provided sim puk is incorrect'
						puk_tries++
						// Send puk, and reset the pin
						await this.exec('AT+CPIN=' + SIM_PUK + ',' + SIM_PIN)
						// Wait 2 secs
						await new Promise(done => setTimeout(done, 2000))
						break
				}
			}

		} catch (err) {

			// Could not unlock sim
			console.log('WARNING: Could not unlock SIM', err)

		}

		mutex.release()
	}

	// Execute a single command
	exec (command) {

		return new Promise((resolve, reject) => {

			this.command = {
				command: command,
				resolve: resolve,
				reject: reject
			}

			if (!this.port.isOpen) {
				// A port nincs is nyitva
				reject('FONA connection is not open')
			}

			// Válasz cache ürítése
			this.cache = []

			// Parancs elküldése
			this.port.write(command + '\r')
		})
	}

	process () {
		console.log('OK')
		this.command.resolve(this.cache)
	}

	error () {
		console.log('ERROR')
		this.command.reject(this.cache)
	}

}
