'use strict';
const CONST  = require("../constants.js");
const LOG    = require("../core/Logger.js");

const ServerMessageBuilder = require("./messages/ServerMessageBuilder.js");

class Channel {
	constructor(callId1, callId2, clientStore) {
		this._id     = Channel.generateChannelId(callId1, callId2);
		this._clientStore = clientStore;
		this._logPrefix = "{" + this._id + "}";

		//making sure callId1 < callId2
		this._callId1 = callId1 < callId2 ? callId1 : callId2;
		this._callId2 = callId1 < callId2 ? callId2 : callId1;

		this._clients = {
			[callId1] : null,
			[callId2] : null
		}
	}

	addClient(client) {
		let id = client.getCallId();
		if (this._clients[id] === undefined) {
			LOG.error(this._logPrefix, "Client with id '" + id + "' is not member of this channel, cannot add!");
			return;
		}

		if (this._clients[id] != null) {
			LOG.warn(this._logPrefix, "Client with id '" + id + "' is already added, updating!");
		}

		LOG.info(this._logPrefix, "Client with id '" + id + "' is added to the channel!")
		this._clients[id] = client;

		if (this._clients[client.getPartnerCallId()]) {
			this._clients[client.getPartnerCallId()].send(ServerMessageBuilder
				.message(ServerMessageBuilder.MESSAGE_TYPE_CONTROL, this)
				.build()
			)
		}
	}

	removeClient(client) {
		let id = client.getCallId();
		if (this._clients[id] === undefined) {
			LOG.error(this._logPrefix, "Client with id '" + id + "' is not member of this channel, cannot remove!");
			return;
		}

		if (this._clients[id] === null) {
			LOG.warn(this._logPrefix, "Client with id '" + id + "' has already been removed from this channel!");
		}

		this._clients[id] = null;

		if (this._clients[this._callId1] === null && this._clients[this._callId2] === null) {
			LOG.info(this._logPrefix, "All clients have been removed from channel, cleaning it up.");
			this._cleanup();
		} else {
			if (this._clients[client.getPartnerCallId()]) {
				this._clients[client.getPartnerCallId()].send(ServerMessageBuilder
					.message(ServerMessageBuilder.MESSAGE_TYPE_CONTROL, this)
					.build()
				)
			}
		}
	}

	sentTo(recipient, message) {
		if (!this._clients[recipient]) {
			LOG.error(this._logPrefix, "Unknown recipient '" + recipient + "' cannot send!");
			return;
		}

		this._clients[recipient].send(message);
	}

	hasClient(callId) {
		return this._clients[callId] ? true : false;
	}

	getConnections() {
		let ret = 0;

		if (this._clients[this._callId1]) ret++;
		if (this._clients[this._callId2]) ret++;

		return ret;
	}

	_cleanup() {
		this._clientStore.removeChannel(this._id);
	}
}

Channel.generateChannelId = function(callId1, callId2) {
	if (callId2 < callId1) {
		let t = callId1;
		callId1 = callId2;
		callId2 = t;
	}

	return callId1 + "-" + callId2;
}

module.exports = Channel;
