'use strict';
const CONST = require("../constants.js");
const LOG   = require("../core/Logger.js");


const $ALL  = Symbol("all");
const $VAR  = Symbol("var");
const $VARLIST = Symbol("vars");
const $PATH = Symbol("path");
const $DATA = Symbol("data");

class Router {
	constructor(handlers) {
		this._handlers     = [];
		this._routingTable = null;

		if (handlers) this.addHandler(handlers);
	}

	addHandler(handlerCls) {
		if (Array.isArray(handlerCls)) {
			this._handlers.push.apply(this._handlers, handlerCls);
		} else {
			this._handlers.push(handlerCls);
		}

		return this;
	}

	build() {
		let routes = {};
		for (let i in this._handlers) {
			routes[this._handlers[i].getRoute()] = this._handlers[i];
			LOG.info("Registered route:", this._handlers[i].getRoute(), "(" + this._handlers[i].name + ")");
		}

		this._routingTable = Router.buildRoutingTable(routes);
	}

	getHandlerData(route) {
		if (!this._routingTable) this.build();
		return Router.matchRoute(route, this._routingTable);
	}
}

Router.prepareRoute = route => {
	//striping leading and trailing slashes
	let left  = route[0] == "/" ? 1 : 0;
	let right = route[route.length - 1] == "/" ? 1 : 0;
	route = route.substr(left, route.length - left - right);

	//splitting
	let parts = route.split("/").filter( s => s.length > 0);
	if (parts.length === 0) {
		parts = [""];
	}

	return parts;
}

Router.buildRoutingTable = routes => {
	let ret = {};
	for (let i in routes) {
		let route     = i;
		let routeData = routes[i];
		let parts     = Router.prepareRoute(i);

		let vars = [];
		let pnt = ret;
		//we will iterate through the whole path and creating the nodes in the table
		for (let j = 0; j < parts.length; j++) {
			let part = parts[j];
			//handling special parts
			if (part === "*")    part=$ALL;
			if (part[0] === ":") {
				vars.push(part.substr(1));
				part=$VAR;
			}

			//creating node for the part
			if (!pnt[part]) pnt[part] = {};
			//our new pointer in the table will be the new node
			pnt = pnt[part];
		}

		//path collision detection
		if (pnt[$PATH]) {
			throw new Error("Path collision: '" + i + "' collides with '" + pnt[$PATH] + "'");
		} else {
			//registering endpoint
			pnt[$DATA] = routeData;
			pnt[$PATH] = i;
			pnt[$VARLIST] = vars;
		}
	}

	return ret;
}

Router.matchRoute = (route, routeTable) => {
	let parts = Router.prepareRoute(route);
	let lasterisk = null;

	let pnt     = routeTable;
	let varsack = [];
	let lasteriskVarsack = [];
	for (let j = 0; j < parts.length; j++) {
		let part = parts[j];
		lasteriskVarsack = [...varsack];
		lasterisk = pnt[$ALL] ? pnt[$ALL] : lasterisk;

		if (pnt[part]) {
			pnt = pnt[part];
		} else if (pnt[$VAR]) {
			varsack.push(part);
			pnt = pnt[$VAR];
		} else {
			pnt     = lasterisk;
			varsack = lasteriskVarsack;
			break;
		}
	}

	if (!pnt || !pnt[$PATH]) {
		pnt     = lasterisk;
		varsack = lasteriskVarsack;
	}

	if (pnt && pnt[$PATH]) {
		let ret = {
			vars : {},
			handler : pnt[$DATA]
		}

		for (let i = 0; i < pnt[$VARLIST].length; i++) ret.vars[pnt[$VARLIST][i]] = varsack[i];
		return ret;
	}

	return null;
}

module.exports = Router;