'use strict';
const CONST = require("../constants.js");
const LOG   = require("../core/Logger.js");

const ServiceBase    = require("../core/ServiceBase.js");
const CacheLRUImpl   = require("../core/CacheLRUImpl.js");
const Crypto         = require("../core/Crypto.js");
const Router         = require("./Router.js");

const MongoConnector         = require("./dal/MongoConnector.js");
const CachingDataAccessLayer = require("./dal/CachingDataAccessLayer.js");

const http        = require("http");
const nodemailer  = require("nodemailer");

class WebAPIService extends ServiceBase {
	_initialize(done) {
		this._port       = this._config.get("api.port", 8888);
		this._pathPrefix = this._config.get("api.pathPrefix", "");
		this._externalHost = this._config.get("api.externalHost", "") + this._pathPrefix;


		LOG.info("Initializing http service on port " + this._port);
		this._server = http.createServer((req, res) => this._connectionHandler(req, res));
		this._server.keepAliveTimeout = this._config.get("api.keep-alive", 1000);
		this._server.timeout = this._config.get("api.timeout", 1000);

		LOG.info("Setting up routing");
		this._router = new Router();
		this._registerHandlers();

		this._setupCache();
		this._setupConnector();
		this._setupDataAccessLayer();
		this._setupCrypto();
		this._setupMail();

		this._requestCnt = 0;
		done();
	}

	_setupCache() {
		this._cache = new CacheLRUImpl(this._config.get("dal.cache_size", 1000));
	}

	_setupConnector() {
		this._connector = new MongoConnector(
			this._config.get("dal.url", "mongodb://localhost"),
			this._config.get("dal.db", "smbui")
		);
	}

	_setupDataAccessLayer() {
		this._dal = new CachingDataAccessLayer(this._connector, this._cache);
	}

	_setupCrypto() {
		this._crypto = new Crypto(this._config.get("auth.crypt", "default_crypto_key"), "hex");
	}

	_setupMail() {
		this._transporter = nodemailer.createTransport({
			host: this._config.get("mail.host", "localhost"),
			port: this._config.get("mail.port", 587),
			secure: this._config.get("mail.secure", false),
			auth: {
				user: this._config.get("mail.auth.user", "noone"),
				pass: this._config.get("mail.auth.pass", "password"),
			},
		});

		this._mail = {
			sendMail : options => this._transporter.sendMail({
				...options,
				from : this._config.get("mail.sender", "noreply@localhost"),
			})
		}
	}

	_registerHandlers() {
		let handlers = [
			require("./handler/CatchAll.js"),
			require("./handler/ActivationHandler.js"),
			require("./handler/AuthHandler.js"),
			require("./handler/UserHandler.js"),
			require("./handler/ResetPasswordHandler.js"),
			require("./handler/UserRegistrationHandler.js")
		];

		if (this._config.get("dev", false)) {
			handlers.push(require("./handler/RestartHandler.js"));
		}

		this._router.addHandler(handlers).build();
	}

	_run() {
		LOG.info("Connecting to database");
		this._connector
			.connect()
			.then(r => {
				LOG.info("Starting the webserver");
				this._server.listen(this._port);
			})
			.catch(e => {
				LOG.error("Error during startup", e);
			});
	}

	_connectionHandler(request, response) {
		LOG.info(request.method, request.url);

		let body = [];
		request.id = ++this._requestCnt;
		request.on('data', (chunk) => {
			body.push(chunk);
		}).on('end', () => {
			request.body = Buffer.concat(body).toString();

			this._route(request)
				.then(handler => this._handleRequest(handler, request))
				.then(out     => this._writeResponse(request.id, out, response))
				.catch(error  => this._fatalErrorResponse(request.id, error, response));
		});
	}

	_route(request) {
		return new Promise((resolve, reject) => {
			let url = request.url;
			if (this._pathPrefix && url.indexOf(this._pathPrefix) === 0) {
				url = url.substr(this._pathPrefix.length);
			}
			request.processedUrl = url;
			let handlerData = this._router.getHandlerData(url);
			if (handlerData === null) throw new Error("No handler found for request: " + request.url);

			LOG.debug("|#" + request.id + "|", request.method, request.url, "Handler:", handlerData.handler.name);
			resolve(handlerData);
		});
	}

	_handleRequest(handlerData, request) {
		let handler = new handlerData.handler(
			request,
			handlerData.vars,
			{
				dal    : this._dal,
				crypto : this._crypto,
				mail   : this._mail,
				externalHost : this._externalHost
			}
		);
		return handler.handle();
	}

	_writeResponse(requestId, out, response) {
		LOG.info("|#" + requestId + "|", "Response", out.code);
		LOG.debug("|#" + requestId + "|", "Final output object", out);

		if (out.headers) {
			for (let i in out.headers) {
				response.setHeader(i, out.headers[i]);
			}
		}
		response.statusCode = out.code;
		response.write(out.body);

		response.end();
		LOG.debug("|#" + requestId + "|", "Transaction end");
	}

	_fatalErrorResponse(requestId, error, response) {
		LOG.error("|#" + requestId + "|", "Unhandled error:", error);
		response.statusCode = 500;
		response.write(JSON.stringify({
			ok      : false,
			error   : {
				reason : "Unhandled server side boo-boo!"
			},
			ver     : CONST.PROTOCOL_VERSION,
			ver_min : CONST.PROTOCOL_COMPATIBLITY_VERSION,
		}));
		response.end();
	}

	_terminate(done) {
		this._server.close();
		this._connector.close();
	}
}

module.exports = WebAPIService;
