'use strict';
const CONST = require("../../constants.js");
const LOG   = require("../../core/Logger.js");

const ConnectorBase = require("./ConnectorBase.js");
const DataLayerException = require("./DataLayerException.js");

const MongoClient = require("mongodb").MongoClient;

class MongoConnector extends ConnectorBase {
	constructor(url, dbName, defaultKeyName) {
		super();
		this._url       = url;
		this._dbName    = dbName;
		this._connected = false;
		this._defaultKeyName = defaultKeyName || "_id";
		this._db        = null;

		this._client = new MongoClient(this._url, {
			useNewUrlParser: true,
			useUnifiedTopology: true
		});

		LOG.info("MongoDB connector created");
	}

	connect() {
		return this._client.connect()
			.then(conn => {
				this._conn = conn;
				return conn.db(this._dbName)
			})
			.then(db => {
				this._db = db;
				this._connected = true;
				LOG.info("MongoDB connector connected to", this._url, "using database", this._dbName);
			});
	}

	close() {
		this._connected = false;
		return this._conn.close();
	}

	isConnected() {
		return this._connected;
	}

	getRecord(table, keyValue, keyName) {
		if (!this.isConnected()) return Promise.reject(new DataLayerException("Not connected to the database!", DataLayerException.ERROR_NOT_CONNECTED));
		keyName = keyName || this._defaultKeyName;
		if (!keyName) return Promise.reject(new DataLayerException("No key name provided!", DataLayerException.ERROR_UNKNOWN_KEY_NAME));
		if (!keyValue) return Promise.reject(new DataLayerException("No key value provided!", DataLayerException.ERROR_MISSING_KEY));

		return new Promise((resolve, reject) => {
			this._db.collection(table).findOne(
				{[keyName] : keyValue},
				(error, result) => {
					if (error) reject(error);
					resolve(result);
				}
			);
		});
	}

	deleteRecord(table, keyValue, keyName) {
		if (!this.isConnected()) return Promise.reject(new DataLayerException("Not connected to the database!", DataLayerException.ERROR_NOT_CONNECTED));
		keyName = keyName || this._defaultKeyName;
		if (!keyName) return Promise.reject(new DataLayerException("No key name provided!", DataLayerException.ERROR_UNKNOWN_KEY_NAME));
		if (!keyValue) return Promise.reject(new DataLayerException("No key value provided!", DataLayerException.ERROR_MISSING_KEY));

		return new Promise((resolve, reject) => {
			this._db.collection(table).deleteOne(
				{[keyName] : keyValue},
				(error, result) => {
					if (error) reject(error);
					resolve(result);
				}
			);
		});
	}

	updateRecord(table, data, keyValue, keyName) {
		if (!this.isConnected()) return Promise.reject(new DataLayerException("Not connected to the database!", DataLayerException.ERROR_NOT_CONNECTED));
		keyName = keyName || this._defaultKeyName;
		keyValue = keyValue || data[keyName];
		if (!keyName) return Promise.reject(new DataLayerException("No key name provided!", DataLayerException.ERROR_UNKNOWN_KEY_NAME));
		if (!keyValue) return Promise.reject(new DataLayerException("No key value provided!", DataLayerException.ERROR_MISSING_KEY));

		return new Promise((resolve, reject) => {
			this._db.collection(table).replaceOne(
				{[keyName] : keyValue},
				{...data, [keyName] : keyValue},
				{},
				(error, result) => {
					if (error) reject(error);
					resolve(result);
			});
		});
	}

	insertRecord(table, data, keyValue, keyName) {
		if (!this.isConnected()) return Promise.reject(new DataLayerException("Not connected to the database!", DataLayerException.ERROR_NOT_CONNECTED));
		keyName = keyName || this._defaultKeyName;
		keyValue = keyValue || data[keyName];
		if (!keyName) return Promise.reject(new DataLayerException("No key name provided!", DataLayerException.ERROR_UNKNOWN_KEY_NAME));
		if (!keyValue) return Promise.reject(new DataLayerException("No key value provided!", DataLayerException.ERROR_MISSING_KEY));

		return new Promise((resolve, reject) => {
			this._db.collection(table).insertOne(
				{...data, [keyName] : keyValue},
				{},
				(error, result) => {
					if (error) reject(error);
					resolve(result);
				}
			);
		});
	}

	saveRecord(table, data, keyValue, keyName) {
		if (!this.isConnected()) return Promise.reject(new DataLayerException("Not connected to the database!", DataLayerException.ERROR_NOT_CONNECTED));
		keyName = keyName || this._defaultKeyName;
		keyValue = keyValue || data[keyName];
		if (!keyName) return Promise.reject(new DataLayerException("No key name provided!", DataLayerException.ERROR_UNKNOWN_KEY_NAME));
		if (!keyValue) return Promise.reject(new DataLayerException("No key value provided!", DataLayerException.ERROR_MISSING_KEY));

		return new Promise((resolve, reject) => {
			this._db.collection(table).replaceOne(
				{[keyName] : keyValue},
				{...data, [keyName] : keyValue},
				{upsert : true},
				(error, result) => {
					if (error) reject(error);
					resolve(result);
			});
		});
	}
}

module.exports = MongoConnector;