'use strict';
const ManagedObjectBase = require("../Components/ManagedObjectBase.js");

const {
	UnimplementedMethodException,
	InvalidArgumentException,
	MissingFieldException
} = require('../Exception/Exceptions.js');

class BusinessLogicBase extends ManagedObjectBase {
	_init() {
		this._actionHandlers = {};

		for (let actionName in this.Actions._HANDLER_MAP) {
			let handlerName = this.Actions._HANDLER_MAP[actionName];

			if (typeof(this[handlerName]) !== "function") throw new UnimplementedMethodException("Action handler method '" + handlerName + "' is not definied in business handler, but action '" + actionName + "' requires it!");
			this._actionHandlers[actionName] = handlerName;
		}
	}

	handle(action, state, dispatch) {
		let handlerName = this._actionHandlers[action.type];

		if (!handlerName) {
			throw new InvalidArgumentException("Unknown action received by business logic handler '" + this.getManagedName() + "' action: '" + action.type + "'");
		}

		if (!Array.isArray(action.payload)) {
			throw new InvalidArgumentException("Action '" + action.type + "' had an invalid payload '" + JSON.stringify(action.payload) + "'");
		}

		var args = [state.getState(), dispatch, ...action.payload];
		this[handlerName].apply(this, args);
	}
}

BusinessLogicBase.generateActions = function(cls, actions, silent) {
	let base_name = cls.prototype.MANAGED_NAME;

	if (!cls.Actions) {
		//making sure the static field is also availalbe through the prototype
		cls.Actions = {};
		cls.prototype.Actions = cls.Actions;

		//setting up internal variables for the business logic registration
		cls.Actions._MANAGED_NAME = base_name;
		cls.Actions._HANDLER_MAP  = {};
	}

	let actualBaseName = "smbfw.businesslogic." + base_name + ".";

	for (let i in actions) {
		//fully qualified smbfw action name
		let actualName  = actualBaseName + i;
		//genarating handler name from the action name if no handler name provided
		let handlerName = actions[i] || (i + "ActionHandler");

		cls.Actions[i] = function() {
			return {
				type    : actualName,
				payload : Array.prototype.slice.call(arguments),
				nolog   : silent
			}
		}

		//generating some information methods and fields for the action
		cls.Actions[i].match = function(otherAction) { return otherAction.type === this.actualName };
		cls.Actions[i].type  = actualName;

		//setting callback map
		cls.Actions._HANDLER_MAP[actualName] = handlerName;
	}

	cls.Actions.toMapDispatchToProps = function(otherActions) {
		var ret = otherActions || {};

		for (let i in cls.Actions) {
			if (typeof cls.Actions[i] === "function") {
				ret[i] = cls.Actions[i];
			}
		}
		return ret;
	};
}

BusinessLogicBase.generateSilentActions = function(cls, actions) {
	return BusinessLogicBase.generateActions(cls, actions, true)
}

//BusinessLogicBase.prototype.MANAGED_NAME = null;
module.exports = BusinessLogicBase;