'use strict';
const DataPipe       = require("./DataPipe.js");
const DataTypeTagger = require("./Middleware/DataTypeTagger.js");
const Util           = require("../Util/Util.js");

const {
	InvalidArgumentException
} = require('../Exception/Exceptions.js');


class DataFactoryBase {
	constructor() {
		this._dataRegistry = {};
		this._setupRegistry();
	}

	_setupRegistry() {}


	registerDataSource(name, provider, middlewares, dataModelClass) {
		if (this._dataRegistry[name]) {
			throw new InvalidArgumentException("Data pipe with name '" + name + "' is already registered");
		}

		this._dataRegistry[name] = {
			pipe  : new DataPipe(provider, middlewares),
			model : dataModelClass || ModelBase
		}
	}

	registerDataPipe(name, pipe, model) {
		if (this._dataRegistry[name]) {
			throw new InvalidArgumentException("Data pipe with name '" + name + "' is already registered");
		}

		this._dataRegistry[name] = {
			pipe  : pipe,
			model : model
		}
	}

	registerDataPipes(pipeDesc) {
		for (let i in pipeDesc) {
			this.registerDataPipe(i, pipeDesc[i].pipe, pipeDesc[i].model);
		}
	}

	getDataPipe(name) {
		if (!this._dataRegistry[name]) {
			throw new InvalidArgumentException("No data pipe found with name '" + name + "'!");
		}

		return this._dataRegistry[name].pipe;
	}

	createRESTBuilder(baseUrl, middlewares, provider) {
		let ret = new RESTBuilder(baseUrl, middlewares, provider);

		ret.registerPipes = () => {
			this.registerDataPipes(ret.buildPipes());
		}

		return ret;
	}

	/**
	 * This method is responsible for loading a data associated to the id provided.
	 * @param {string} name - The name of the data source to be used to create the data
	 * @param {string} id - The key of the data to be loaded
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], where the data is the freshly queried data
	 */
	loadData(name, id, ...vars) {
		if (Array.isArray(vars[0])) vars = vars[0];
		let options = vars[0] === undefined ? undefined : {
			pathParts : vars
		}

		try {
			let pipe = this.getDataPipe(name);
			return pipe.loadData(id, options).then(result => {
				let [id, data, options] = result;
				let model = new this._dataRegistry[name].model(data, pipe, this, options);

				return model;
			});
		} catch (e) {
			return Util.createErrorPromise(e);
		}
	}

	/**
	 * This method is responsible for creating non-existing data.
	 * @param {string} name - The name of the data source to be used to create the data
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], the id will be the created key for the record
	 */
	createData(name, data, ...vars) {
		if (Array.isArray(vars[0])) vars = vars[0];
		let options = vars[0] === undefined ? undefined : {
			pathParts : vars
		}
		let pipe = this.getDataPipe(name);
		let model = new this._dataRegistry[name].model(data || null, pipe, this, options);

		return model;
	}
}

module.exports = DataFactoryBase;