'use strict';
const Util = require('../Util/Util.js');
/*
	Consumer   --request-->   Middleware   --request-->   Middleware   --request-->   Provider
	    ^                                                                                |
	    +--------response--   Middleware  <--response--   Middleware   <--response-------+
	                      /
	  Sink  <--response--/
*/

class DataPipe {
	/**
	 * Constructor
	 * @param {ProviderBase} provider - The provider which will connect to the actual data store
	 * @param {array of objects} middlewares - Array of objects which contain pre- and/or post- middleware methods.
	 */
	constructor(provider, middlewares) {
		this._provider    = provider;
		this._middlewares = [];

		if (middlewares) {
			this._middlewares = Array.isArray(middlewares) ? middlewares : [middlewares];
		}
	}

	/**
	 * This internal helper method will create a promise which will call all the pre-middlewares before calling the providers corresponting method
	 * Also calling the post-middlewares when the response is provided by the provider
	 * @param {array} args - The arguments of the actual exposed method we are going to wrap
	 * @param {string} pipeMethod - The actual method's name, this method will be invoked on the provider, and this method's pre and post methods will be invoked in the middlewares
	 * @return {Promise} - The promise's resolve will provide the processed response from the provider
	 */
	_pipeRequest(args, pipeMethod) {
		//generating pre and post method names based on the pipe method name
		let preMethod  = "pre" + Util.upperCaseInitial(pipeMethod);
		let postMethod = "post" + Util.upperCaseInitial(pipeMethod);

		//creating an error handling promise
		return new Promise((resolve, reject) => {
			//chaining middlewares in order calling the pre methods
			for (let i = 0; i < this._middlewares.length; i++) {
				let mw = this._middlewares[i];
				//we chain the args through middlewares which implement the preMethod
				if (typeof(mw[preMethod]) === "function") args = mw[preMethod].apply(mw, args);
			}

			//after we have the modified argument list we call the provider's method
			this._provider[pipeMethod].apply(this._provider, args).then((result) => {
				//this handler will filter the response through the middlewares in reverse order
				for (let i = this._middlewares.length - 1; i >= 0; i--) {
					let mw = this._middlewares[i];
				//we chain the args through middlewares which implement the postMethod
					if (typeof(mw[postMethod]) === "function") result = mw[postMethod].apply(mw, result);
				}
				//if everything went fine, we call the resolve with our latest result
				resolve(result);
			}).catch(reason => {
				//in the case of an error we simply reject
				reject(reason);
			});
		});
	}

	/**
	 * This method is responsible for updating the data associated to the provided id.
	 * @param {string} id - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options]
	 */
	saveData(id, data, options) {
		return this._pipeRequest([id, data, options || {}], "saveData");
	}

	/**
	 * This method is responsible for loading a data associated to the id provided.
	 * @param {string} id - The key of the data to be loaded
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], where the data is the freshly queried data
	 */
	loadData(id, options) {
		return this._pipeRequest([id, options || {}], "loadData");
	}

	/**
	 * This method is responsible for creating non-existing data.
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], the id will be the created key for the record
	 */
	createData(data, options) {
		return this._pipeRequest([data, options || {}], "createData");
	}

	/**
	 * This method is responsible for listing, the filter should contain all filter specific data, there is no queryu language for that, so it will depend on the implementation anyway.
	 * @param {mixed} filter - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @return {Promise} - The promise's resolve will provide and array consisting of [dataMap, token], where the dataMap is an id -> data map of the items
	 * 						the token is a continuation token, null if lising has finished
	 */
	listData(filter, options) {
		return this._pipeRequest([filter, options || {}], "listData");
	}


	/**
	 * This method is responsible for continue a multi-part listing, the token provided should be the token received by the listData's response
	 * @param {mixed} token - The continuation token which will help the provider to continue the listing
	 * @return {Promise} - The promise's resolve will provide and array consisting of [dataMap, token], where the dataMap is an id -> data map of the items
	 * 						the token is a continuation token, null if lising has finished
	 */
	continueList(token) {
		return this._pipeRequest([token], "continueList");
	}

	/**
	 * This method will iterate through a list, calling the callback for each item.
	 * @param {function(key, data)} fn - the function to be called for each record
	 * @param {mixed} filter - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @return Promise - Fulfilled when list iteration is done, on error catch will be invoked, so it should be handled.
	 */
	listIterate(fn, filter, options) {
		if (typeof fn !== "function") return Util.createErrorPromise("You forgot to provide a callback FUNCTION for listIterate!");

		const iterateList = (res) => {
			let [dataList, token] = res;
			for (let i in dataList) { fn(i, dataList[i]) };
			if (token) return this.continueList(token).then(iterateList);
		}

		return this.listData(filter, options).then(iterateList);
	}
}

module.exports = DataPipe;