'use strict';
const ProviderBase  = require("./ProviderBase.js");
const DataPipeError = require("../DataPipeError.js");
const Util          = require("../../Util/Util.js");

class MemoryProvider extends ProviderBase {
	constructor(initialData, delay, listSize) {
		super();
		this._data     = initialData || {};
		this._delay    = delay || 0;
		this._listSize = listSize || 10;

		this._seq   = 0;
		for (let i in this._data) {
			for (let j in this._data[i]) {
				if (j > this._seq) this._seq = j + 1;
			}
		}
	}

	/**
	 * This method is responsible for updating the data associated to the provided id.
	 * @param {string} id - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options]
	 */
	saveData(id, data, options) {
		let dataType = options.dataType || "_default";
		return new Promise((resolve, reject) => {
			if (id > this._seq) this._seq = id + 1;
			if (!this._data[dataType]) this._data[dataType] = {};
			this._data[dataType][id] = Util.unsafeClone(data);
			setTimeout(() => resolve([id, data, options]), this._delay) ;
		});
	}

	/**
	 * This method is responsible for loading a data associated to the id provided.
	 * @param {string} id - The key of the data to be loaded
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], where the data is the freshly queried data
	 */
	loadData(id, options) {
		let dataType = options.dataType || "_default";
		return new Promise((resolve, reject) => {
			if (this._data[dataType] && this._data[dataType][id]) {
				setTimeout(() => resolve([id, Util.unsafeClone(this._data[dataType][id]), options]), this._delay);
			} else {
				reject(DataPipeError.notFound("ID '" + id + "' is not found in the store", {id : id}));
			};
		});
	}

	/**
	 * This method is responsible for creating non-existing data.
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @param {object} options - Custom options for the provider, since it is connection dependent, middlewares should fill it, to make the data consumer implementation independent
	 * @return {Promise} - The promise's resolve will provide and array consisting of [id, data, options], the id will be the created key for the record
	 */
	createData(data, options) {
		let dataType = options.dataType || "_default";
		return new Promise((resolve, reject) => {
			let id = data.id || options.id || (this._seq++).toString();
			if (!this._data[dataType]) this._data[dataType] = {};
			if (this._data[dataType][id] !== undefined) reject(DataPipeError.alreadyExists("ID '" + id + "' already exists found in the store", {id : id}));
			this._data[dataType][id] = Util.unsafeClone(data);
			setTimeout(() => resolve([id, data, options]), this._delay) ;
		});
	}

	/**
	 * This method is responsible for listing, the filter should contain all filter specific data, there is no queryu language for that, so it will depend on the implementation anyway.
	 * @param {mixed} filter - The key of the data to be saved
	 * @param {mixed} data - The actual data, it should be converted in the middleware layer to the proper format
	 * @return {Promise} - The promise's resolve will provide and array consisting of [dataMap, token], where the dataMap is an id -> data map of the items
	 * 						the token is a continuation token, null if lising has finished
	 */
	listData(filter, options) {
		let dataType = options.dataType || "_default";
		var order = [];
		for (let i in this._data[dataType] || {}) {
			order.push(i);
		}

		let token = {
			order : order,
			dataType : dataType,
			idx   : 0
		}

		return this.continueList(token);
	}

	/**
	 * This method is responsible for continue a multi-part listing, the token provided should be the token received by the listData's response
	 * @param {mixed} token - The continuation token which will help the provider to continue the listing
	 * @return {Promise} - The promise's resolve will provide and array consisting of [dataMap, token], where the dataMap is an id -> data map of the items
	 * 						the token is a continuation token, null if lising has finished
	 */
	continueList(token) {
		let dataType = token.dataType;
		return new Promise((resolve, reject) => {
			let start = token.idx;
			let end   = Math.min(token.idx + this._listSize - 1, token.order.length - 1);
			let ret   = {};
			token.idx = end + 1;

			for (let i = start; i <= end; i++) {
				ret[token.order[i]] = this._data[dataType][token.order[i]];
			}

			if (token.idx >= token.order.length) token = null;
			setTimeout(() => resolve([ret, token]), this._delay);
		})
	}
}

module.exports = MemoryProvider;