'use strict';
const ManagerBase        = require("./ManagerBase.js");
const ScreenBase         = require("../Components/ScreenBase");
const ModalWindowBase    = require("../Components/ModalWindowBase.js");
const React              = require('react');
const Application        = require("../Components/Application.js");
const SmbfwActions    = require("../Action/SmbfwActions.js");

const {
	InvalidStateException,
	InvalidArgumentException,
	InvalidStateChangeException
} = require('../Exception/Exceptions.js');

const MAX_SCREEN_FALLBACK_COUNT = 20;

class ScreenManager extends ManagerBase {
	constructor() {
		super();

		this._currentScreen = null;
		this._initialScreen = null;
		this._screenAvailability = null;
		this._modalWindows  = {};
	}

	_init() {
	}

	registerScreen() {
		for (let i = 0; i < arguments.length; i++) {
			let screen = this._addManaged(arguments[i]);
			if (!this._initialScreen) this.setInitialScreen(screen.getManagedName());
		}
	}

	registerModal() {
		for (let i = 0; i < arguments.length; i++) {
			let modal = this._addModal(arguments[i]);
		}
	}

	_addModal(modalClass) {
		if (modalClass === null || modalClass === undefined) throw new InvalidArgumentException("The registerScreen method expects a CLASS derived from " + this.util.getClassname(ModalWindowBase));
		if (!(modalClass.prototype instanceof ModalWindowBase)) throw new InvalidArgumentException("The registerScreen method expects a CLASS derived from " + this.util.getClassname(ModalWindowBase) + ". Instead it received: " + this.util.getClassname(ModalWindowBase));

		var object = this._smbfw.createSmbfwObject.apply(this._smbfw, arguments);
		var name = object.getManagedName();

		if (this._modalWindows[name]) {
			this._warn("Object with key '" + name + "' is already registered.");
			return this._modalWindows[name];
		}

		this._info("Registering '" + name + "'");
		this._modalWindows[name] = object;
		object._setManager(this);

		return object;
	}

	setInitialScreen(screenId) {
		var screen = this._getManaged(screenId);
		if (screen == null) throw new InvalidArgumentException("No screen found with name '" + screenId + "'");

		this._initialScreen = screen;
		this._info("Initial screen updated to:", this._initialScreen.getManagedName());
	}

	transition(screenName, params) {
		var screen = this._getManaged(screenName);
		this._log("Changing screen to:", screen.getName());

		let update = this._checkScreenAvailability(screen, this._screenAvailability);

		let deep = this._checkDeeplinking(this._screenAvailability);
		if (deep) {
			update = deep;
			this.clearDeeplinking();
		}

		this._currentScreen = update["smbfw.currentScreen"] ? update["smbfw.currentScreen"] : screen;
		this._dispatch(SmbfwActions.updateState({
			"smbfw.screenParams"  : params,
			"smbfw.currentScreen" : this._currentScreen,
			...update
		}));
	}

	showModal(modalName, params) {
		var modal = this._modalWindows[modalName];
		this._log("Displaying modal window:", modal.getName());

		this._currentModal = modal;
		this._dispatch(SmbfwActions.updateState({
			"smbfw.modalParams"  : params,
			"smbfw.modalWindow"  : this._currentModal
		}));
	}

	setDeeplinking(screen, params) {
		this._log("Setting deeplinked screen:", screen);
		this._deepScreen = this._getManaged(screen);
		this._deepParams = params;
	}

	clearDeeplinking() {
		this._deepScreen = null;
		this._deepParams = null;
	}

	hideModal() {
		this._log("Removing any open modal window");
		this._currentModal = null;

		this._dispatch(SmbfwActions.updateState({
			"smbfw.modalParams"  : undefined,
			"smbfw.modalWindow"  : undefined
		}));
	}

	_checkDeeplinking(availability) {
		if (!this._deepScreen) return;
		if (!this._deepScreen.checkScreenAvailability(availability)) return;

		let ret = {
			"smbfw.currentScreen" : this._deepScreen,
			"smbfw.screenParams"  : this._deepParams
		};

		return ret;
	}

	_checkScreenAvailability(screen, availability) {
		availability = availability || "";
		let screenPnt = screen;
		let ret      = {
			"smbfw.screenAvailability" : availability
		};

		let cnt = 0;
		while (!screenPnt.checkScreenAvailability(availability) && screenPnt !== this._initialScreen) {
			let screenName = screenPnt.getFallbackScreen(availability);

			if (!screenName) {
				screenPnt = this._initialScreen;
				break;
			}

			screenPnt = this._getManaged(screenName);
			cnt++;
			//to prevent infinite loops
			if (cnt > MAX_SCREEN_FALLBACK_COUNT) throw new InvalidStateChangeException("Cannot change screen availability, no fallback screen found in " + MAX_SCREEN_FALLBACK_COUNT + " steps!");
		}

		if (screenPnt !== screen) {
			this._log("Screen '" + screen.getName() + "'is no longer available using fallback screen:", screenPnt.getName());
			ret["smbfw.currentScreen"] = screenPnt;
			//we make sure we clear any screen params
			ret["smbfw.screenParams"]  = undefined;
		}

		return ret;
	}

	updateScreenAvailability(availability) {
		let update               = this._checkScreenAvailability(this.getCurrentScreen(), availability);
		this._screenAvailability = availability;
		this._dispatch(SmbfwActions.updateState(update));
	}

	serviceActionHandler(action, next, state) {
		if (SmbfwActions.changeScreen.match(action)) {
			this.transition(action.payload, action.screenParams);
		} else if (SmbfwActions.showModal.match(action)) {
			this.showModal(action.payload, action.modalParams)
		} else if (SmbfwActions.updateScreenAvailability.match(action)) {
			this.updateScreenAvailability(action.payload);
		} else if (SmbfwActions.hideModal.match(action)) {
			this.hideModal();
		} else if (SmbfwActions.setScreenDeeplinking.match(action)) {
			this.setDeeplinking(action.payload, action.params);
		} else {
			next(action);
		}
	}

	start() {
		this.transition(this._initialScreen.getName());
	}

	getCurrentScreen() {
		return this._currentScreen;
	}

	getApplication() {
		return React.createElement(Application, null, null)
	}
}

ScreenManager.prototype.PUBLISHED_METHODS  = ["registerScreen", "setInitialScreen", "registerModal"];
ScreenManager.prototype.MANAGED_BASE_CLASS = ScreenBase;
module.exports = ScreenManager;