'use strict'
const URL = require('url');
const {InvalidArgumentException} = require("../Exception/Exceptions.js");
const Util = {};

Util.getClassname = function(what) {
	if (what === null) return "null";
	if (what === undefined) return "undefined";

	var proto = (typeof what === "function") ? what.prototype : Object.getPrototypeOf(what);
	return (proto && proto.constructor && proto.constructor.name) ? proto.constructor.name : "unknown";
}

Util.createErrorPromise = function(msg) {
	return new Promise((resolve, reject) => {
		reject(msg);
	});
}

Util.createRetryPromise = function(fn, count, delay) {
	delay = delay || 1;
	count = count || 1;
	//this is a wrapper promise which is responsible sending the inner promise multiple times
	return new Promise((resolve, reject) => {
		const createAttempt = () => {
			//each attempt created reduces the number of attempts to try
			count--;
			let attempt = new Promise((res, rej) => {
				let ret = fn(res, rej);
				if (ret instanceof Promise) ret.then( r => res(r)).catch( e => rej(e));
			});

			attempt
				.then(res => resolve(res))
				.catch(reason => {
					//if we are out of retries we reject with the last error
					if (count <= 0) {
						reject(reason);
						return;
					}

					//rescheduling the retry in delay number of ms'
					setTimeout(createAttempt, delay);
				});
			return attempt;
		}

		//creating the first attempt with no delay
		createAttempt();
	});
}

Util.namedPromiseAll = function(promises) {
	var promiseOrder = Object.keys(promises);
	var promiseArray = promiseOrder.map(v => promises[v]);

	return Promise.all(promiseArray).then(res => {
		var ret = {};
		for (let i in res) {
			ret[promiseOrder[i]] = res[i]
		}

		return ret;
	});
}

Util.generateStateChangeTree = function(items) {
	var changeMap = {};
	for (let i in items) {
		let parts = i.split(".");
		let pnt   = changeMap;
		for (let j = 0; j < parts.length - 1; j++) {
			if (!pnt[parts[j]]) {
				pnt[parts[j]] = {};
			}
			pnt = pnt[parts[j]];
			if (typeof pnt !== "object") throw new InvalidArgumentException("Current key " + i + "  is in conflict with an other previously definied key.");
		}

		let lastPart = parts.pop();
		if (typeof pnt[lastPart] !== "undefined") throw new InvalidArgumentException("Current key " + i + "  is in conflict with an other previously definied key.");
		pnt[lastPart] = i;
	}

	return changeMap;
}

Util.applyStateChangeTree = function(source, changeMap, changeList) {
	var ret = {...source};

	for (let i in changeMap) {
		if (typeof changeMap[i] === "string") {
			if (changeList[changeMap[i]] !== undefined) {
				ret[i] = changeList[changeMap[i]];
			} else {
				delete ret[i];
			}
		} else {
			ret[i] = Util.applyStateChangeTree(source[i] || {}, changeMap[i], changeList)
		}
	}

	return ret;
}

Util.createPathPattern = function(...parts) {
	let url       = URL.parse(parts.join("/"));
	let pathParts = url.pathname.split("/");

	let pp      = [];
	let vars    = [];
	let partial = [url.protocol + "//" + url.host];
	for (let i = 0; i < pathParts.length; i++) {
		let part = pathParts[i];

		if (part[0] == ":") {
			if (partial.length) pp.push(partial.join("/"));
			pp.push(":");
			vars.push(pp.length - 1);
			partial = [];
		} else if (part) {
			partial.push(part);
		}
	}

	if (partial.length) pp.push(partial.join("/"));

	return {
		parts : pp,
		vars  : vars
	};
}

Util.substitutePathPattern = function(pattern, ...vars) {
	if (Array.isArray(vars[0])) vars = vars[0];
	if (vars.length !== pattern.vars.length) throw new Error("Pattern expects " + pattern.vars.length + " variables but received " + vars.length);
	let ret = [...pattern.parts];
	for (let i in vars) {
		ret[ pattern.vars[i] ] = vars[i];
	}
	return ret.join("/");
}

Util.resolveReference = function(ref, object, def) {
	let parts = ref.split(".");
	let pnt   = object;
	for (let i = 0; i < parts.length; i++) {
		pnt = pnt[parts[i]];
		if (pnt === undefined || pnt === null) return def;
	}

	return (pnt === undefined) ? def : pnt;
}

Util.resolveDoubleReference = function(ref, object, def) {
	let path = Util.resolveReference(ref, object, "");
	if (typeof path !== "string") return def;
	return Util.resolveReference(path, object, def);
}

Util.unsafeClone = function(o) {
	if (typeof(o) !== "object") return o;
	let ret = Array.isArray(o) ? [] : {};
	for (let i in o) {
		if (o[i] === null) {
			ret[i] = null;
		} else if (typeof o[i] === "object") {
			ret[i] = Util.unsafeClone(o[i]);
		} else {
			ret[i] = o[i];
		}
	}

	return ret;
}

Util.quickHash = function(s) {
	if (typeof s !== "string") {
		return 0;
	}

	var hash = 0, char;
	if (s.length == 0) return hash;

	for (let i = 0, l = s.length; i < l; i++) {
		char = s.charCodeAt(i);
		hash = ((hash << 5) - hash) + char;
		hash |= 0; // Convert to 32bit integer
	}
	return hash;
}

Util.upperCaseInitial = (str => str ? str[0].toUpperCase() + str.substr(1) : str)

module.exports = Util;